(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2010-2017)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2012-2017)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2016)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2013, 2014)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2012)
Author:  Akihisa Yamada <akihisa.yamada@uibk.ac.at> (2017)
License: LGPL (see file COPYING.LESSER)
*)
theory CPF_Parser
imports
  Proof_Checker
  XTC_Parser
  LTS.LTS_Parser
begin

hide_const (open) Congruence.partition
hide_const (open) Invariant_Checker.name
hide_const (open) AC_Subterm_Criterion.proj_term
hide_const (open) Finite_Cartesian_Product.mat
hide_const (open) Simplex.Bounds
hide_const (open) Cooperation_Program.sharp.Sharp
hide_type (open) Finite_Cartesian_Product.vec
hide_const (open) eq

(**** not yet adapted to the new XML parser ****)

partial_function (sum_bot)
  arith_fun :: "arithFun xmlt"
where
  [code]: "arith_fun xml = Xmlt.singleton ''arithFunction'' (Xmlt.options [
     (''natural'', Xmlt.change (Xmlt.nat ''natural'') Const),
     (''variable'', Xmlt.change (Xmlt.nat ''variable'') (\<lambda>n::nat. Arg (n - 1))),
     (''sum'', Xmlt.many ''sum'' arith_fun Sum),
     (''product'', Xmlt.many ''product'' arith_fun Prod),
     (''min'', Xmlt.many ''min'' arith_fun Min),
     (''max'', Xmlt.many ''max'' arith_fun Semantic_Labeling_Carrier.Max),
     (''ifEqual'', Xmlt.tuple4 ''ifEqual'' arith_fun arith_fun arith_fun arith_fun IfEqual)
   ]) id  xml"

definition
  sl_variant :: "'f xmlt \<Rightarrow> ('f,string) sl_variant xmlt"
where
  "sl_variant xml2name \<equiv> Xmlt.singleton ''model'' (Xmlt.options [
     (''finiteModel'', xml1or2many_elements ''finiteModel''
         (Xmlt.change (Xmlt.nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
         (Xmlt.singleton ''tupleOrder'' (Xmlt.leaf ''pointWise'' ()) id)
         (Xmlt.triple ''interpret''
           xml2name
           (Xmlt.nat ''arity'')
           CPF_Parser.arith_fun
           (\<lambda>f a inter. ((f, a), inter)))
         (\<lambda>n model xs. if model = None then Finitelab (SL_Inter n xs) else QuasiFinitelab (SL_Inter n xs) [])),
     (''rootLabeling'',Xmlt.guard (\<lambda> x. Xml.num_children x = 0)
            (Xmlt.leaf ''rootLabeling'' (Rootlab None))
            (Xmlt.singleton ''rootLabeling'' xml2name (\<lambda> f. Rootlab (Some (f,1))))
        )
   ]) id"
hide_const (open) sl_variant

context
  fixes C :: "'a ring" (structure)
begin
fun sum_lpoly :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "sum_lpoly [] ys = ys"
| "sum_lpoly xs [] = xs"
| "sum_lpoly (x # xs) (y # ys) = (x \<oplus> y) # sum_lpoly xs ys"

fun
  lpoly_of ::
    "(nat, 'a) tpoly \<Rightarrow> string +\<^sub>\<bottom> ('a \<times> 'a list)"
where
  "lpoly_of (PNum i) = return (i, [])"
| "lpoly_of (PVar x) = return (\<zero>, replicate x \<zero> @ [\<one>])"
| "lpoly_of (PSum []) = return (\<zero>, [])"
| "lpoly_of (PSum (p # ps)) = (
     lpoly_of  p
  \<bind> (\<lambda>(cp, ncp). lpoly_of (PSum ps)
  \<bind> (\<lambda>(cq, ncq). return (cp \<oplus> cq, sum_lpoly ncp ncq))))"
| "lpoly_of (PMult []) = return (\<one>, [])"
| "lpoly_of (PMult (p # ps)) = (
     lpoly_of p
  \<bind> (\<lambda>(cp, ncp). lpoly_of (PMult ps)
  \<bind> (\<lambda>(cq, ncq). if (Ball (set ncp) ((=) \<zero>))
       then return (cp \<otimes> cq, map ((\<otimes>) cp) ncq)
       else (if (Ball (set ncq) ((=) \<zero>))
         then return (cp \<otimes> cq, map (\<lambda> x. x \<otimes> cq) ncp)
         else error ''cannot transform non-linear polynomial to linear polynomial''))))"

primrec fit_length :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "fit_length (Suc n) as = (case as of
    [] \<Rightarrow> replicate (Suc n) \<zero>
  | b # bs \<Rightarrow> b # fit_length n bs)"
| "fit_length 0 _ = []"
end

partial_function (sum_bot) polynomial :: "'a xmlt \<Rightarrow> (nat, 'a) tpoly xmlt" where
  [code]: "polynomial xml2coeff x = Xmlt.singleton ''polynomial'' (Xmlt.options [
    (''coefficient'', Xmlt.singleton ''coefficient'' xml2coeff PNum),
    (''variable'', Xmlt.change (Xmlt.nat ''variable'') (\<lambda>n. PVar (n - 1))),
    (''sum'', Xmlt.many ''sum'' (polynomial xml2coeff) PSum),
    (''product'', Xmlt.many ''product'' (polynomial xml2coeff) PMult)
  ]) id x"

definition int_coeff :: "int xmlt" where
  "int_coeff \<equiv> Xmlt.int ''integer''"
hide_const (open) int_coeff

definition vec_coeff :: "'a xmlt \<Rightarrow> 'a vec xmlt" where
  "vec_coeff xml2coeff \<equiv> Xmlt.many ''vector''
     (Xmlt.singleton ''coefficient'' xml2coeff id) vec_of_list"
hide_const (open) vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition mat_coeff :: "nat \<Rightarrow> 'a \<Rightarrow> 'a xmlt \<Rightarrow> 'a mat xmlt" where
  "mat_coeff n ze xml2coeff \<equiv> Xmlt.options [
     (''matrix'', Xmlt.many ''matrix'' (CPF_Parser.vec_coeff xml2coeff) (mat_of_cols n)),
     (''vector'', Xmlt.change
       (CPF_Parser.vec_coeff xml2coeff)
       (\<lambda>v. mat n n (\<lambda> (i,j). if j = 0 then v $ i else ze))
     )]"
hide_const (open) mat_coeff

definition arctic_coeff :: "arctic xmlt" where
  "arctic_coeff \<equiv> Xmlt.options [
    (''integer'', Xmlt.change (Xmlt.int ''integer'') Num_arc),
    (''minusInfinity'', Xmlt.leaf ''minusInfinity'' MinInfty)
  ]"
hide_const (open) arctic_coeff

definition arctic_rat_coeff :: "rat arctic_delta xmlt" where
  "arctic_rat_coeff \<equiv> Xmlt.options [
    (''integer'', Xmlt.change Xmlt.rat Num_arc_delta),
    (''rational'', Xmlt.change Xmlt.rat Num_arc_delta),
    (''minusInfinity'', Xmlt.leaf ''minusInfinity'' MinInfty_delta)
  ]"
hide_const (open) arctic_rat_coeff

datatype Domain =
  Natural nat | Integer | Arctic | Arctic_rat | Int_mat nat nat | Arctic_mat nat
| Arctic_rat_mat nat | Rational rat nat | Rat_mat nat nat | Mini_Alg real nat
| Mini_Alg_mat nat nat

definition rat_domain :: "rat xmlt" where
  "rat_domain \<equiv> Xmlt.singleton ''rationals'' (Xmlt.singleton ''delta'' Xmlt.rat id) id"
hide_const (open) rat_domain

definition real :: "real xmlt" where
  "real x = (try
    Xmlt.change Xmlt.rat (real_of_rat) x
  catch (\<lambda>_.
    Xmlt.triple ''algebraic'' (Xmlt.rat) (Xmlt.rat) (Xmlt.rat)
      (\<lambda>a b c. (real_of_rat a + real_of_rat b * sqrt (real_of_rat c))) x))"

definition real_domain :: "real xmlt" where
  "real_domain = Xmlt.singleton ''algebraicNumbers'' (Xmlt.singleton ''delta'' real id) id"
hide_const (open) real_domain

definition basic_domain :: "(nat \<Rightarrow> Domain) xmlt" where
  "basic_domain \<equiv> Xmlt.options [
    (''naturals'', Xmlt.leaf ''naturals'' Natural),
    (''integers'', Xmlt.leaf ''integers'' (\<lambda> d. Integer)),
    (''arctic'', Xmlt.singleton ''arctic''
      (Xmlt.singleton ''domain'' (Xmlt.options [
        (''naturals'', Xmlt.leaf ''naturals'' (\<lambda> d. Arctic)),
        (''integers'', Xmlt.leaf ''integers'' (\<lambda> d. Arctic)),
        (''rationals'', Xmlt.change CPF_Parser.rat_domain (\<lambda> r d. Arctic_rat))
      ]) id) id),
    (''rationals'', Xmlt.change CPF_Parser.rat_domain Rational),
    (''algebraicNumbers'', Xmlt.change CPF_Parser.real_domain Mini_Alg)
  ]"
hide_const (open) basic_domain

definition interpretation_type :: "Domain xmlt" where
  "interpretation_type \<equiv> (Xmlt.singleton ''type''
     (Xmlt.options [
       (''polynomial'', (Xmlt.pair ''polynomial''
       (Xmlt.singleton ''domain''
       (Xmlt.options [
         (''naturals'', CPF_Parser.basic_domain),
         (''integers'', CPF_Parser.basic_domain),
         (''arctic'', CPF_Parser.basic_domain),
         (''rationals'', CPF_Parser.basic_domain),
         (''algebraicNumbers'', CPF_Parser.basic_domain),
         (''matrices'',Xmlt.triple ''matrices''
           (Xmlt.nat ''dimension'')
           (Xmlt.nat ''strictDimension'')
           (Xmlt.singleton ''domain'' CPF_Parser.basic_domain id)
           (\<lambda>di sd domain d. case domain d of
             Natural _ \<Rightarrow> Int_mat di sd
           | Rational _ _ \<Rightarrow> Rat_mat di sd
           | Mini_Alg _ _ \<Rightarrow> Mini_Alg_mat di sd
           | Integer \<Rightarrow> Int_mat di sd
           | Arctic \<Rightarrow> Arctic_mat di
           | Arctic_rat \<Rightarrow> Arctic_rat_mat di
         ))
       ])
     id)
     (Xmlt.nat ''degree'')
     (\<lambda>type d. type d))),
     (''matrixInterpretation'',Xmlt.triple ''matrixInterpretation''
       (Xmlt.singleton ''domain'' CPF_Parser.basic_domain id)
       (Xmlt.nat ''dimension'')
       (Xmlt.nat ''strictDimension'')
       (\<lambda>domain di sd. case domain 0 of
         Natural _ \<Rightarrow> Int_mat di sd
       | Integer \<Rightarrow> Int_mat di sd
       | Rational _ _  \<Rightarrow> Rat_mat di sd
       | Mini_Alg _ _  \<Rightarrow> Mini_Alg_mat di sd
       | Arctic \<Rightarrow> Arctic_mat di
       | Arctic_rat \<Rightarrow> Arctic_rat_mat di
     ))
   ])
 id)"
hide_const (open) interpretation_type

type_synonym ('f,'a) inter = "('f \<times> nat) \<times> 'a \<times> 'a list"
type_synonym ('f,'a) mat_inter = "('f \<times> nat) \<times> 'a mat \<times> 'a mat list"
type_synonym ('f,'a) nl_inter = "('f \<times> nat) \<times> (nat, 'a) poly"

datatype (dead 'f) "interpretation" =
  (* linear polynomials (with support of negative constants *)
  Int_linear_poly (int_linear_poly: "('f,int)inter")
| Rat_linear_poly  (rat_linear_poly: "('f,rat)inter")
| Arctic_linear_poly (arctic_linear_poly:  "('f,arctic)inter")
| Arctic_rat_linear_poly  (arctic_rat_linear_poly: "('f,rat arctic_delta)inter")
| Real_linear_poly  (real_linear_poly: "('f,real)inter")
  (* linear polynomials over matrices (with support of negative constants *)
| Int_matrix  (int_matrix: "('f,int)mat_inter")
| Rat_matrix  (rat_matrix: "('f,rat)mat_inter")
| Arctic_matrix  (arctic_matrix: "('f,arctic)mat_inter")
| Arctic_rat_matrix  (arctic_rat_matrix: "('f,rat arctic_delta)mat_inter")
| Real_matrix  (real_matrix: "('f,real)mat_inter")
  (* non linear polynomials (without support of negative constants) *)
| Int_non_linear_poly  (int_non_linear_poly: "('f,int)nl_inter")
| Rat_non_linear_poly  (rat_non_linear_poly: "('f,rat)nl_inter")
| Real_non_linear_poly  (real_non_linear_poly: "('f,real)nl_inter")

(* the flag bi is to indicate that we are interested in bounded increase,
   where rationals always have to be chosen as non-linear variant *)
abbreviation class_ring where
  "class_ring t \<equiv> class_semiring t ()"

abbreviation mat_ring :: "('a :: semiring_1) itself \<Rightarrow> nat \<Rightarrow> 'a mat ring" where
  "mat_ring t n \<equiv> ring_mat t n ()"
 
definition "interpretation" :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "interpretation bi xml2name \<equiv>
      (Xmlt.many1_gen ''interpretation''
       CPF_Parser.interpretation_type
       (\<lambda>type. case type of
          Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Int_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                   (class_ring (TYPE(int))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<bind> (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Int_non_linear_poly ((f,a),poly))))
        | Integer \<Rightarrow> Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<bind> (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Int_non_linear_poly ((f,a),poly)))
        | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xmlt.rat x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (class_ring (TYPE(rat))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xmlt.rat x \<bind> (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Rat_non_linear_poly ((f,a),poly))))
        | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial real x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Real_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (class_ring (TYPE(real))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial real x \<bind> (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Real_non_linear_poly ((f,a),poly))))
        | Arctic \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_coeff x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Arctic_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (class_ring (TYPE(arctic)))
        | Arctic_rat \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_rat_coeff x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Arctic_rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (class_ring (TYPE(rat arctic_delta)))
        | Int_mat n sd \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n 0 CPF_Parser.int_coeff) x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Int_matrix ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (mat_ring (TYPE(int)) n)
        | Rat_mat n sd \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n 0 Xmlt.rat) x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Rat_matrix ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (mat_ring (TYPE(rat)) n)
        | Mini_Alg_mat n sd \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n 0 real) x \<bind> lpoly_of c)
                        (\<lambda>f a poly.  (Real_matrix ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (mat_ring (TYPE(real)) n)
        | Arctic_mat n \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n MinInfty CPF_Parser.arctic_coeff) x \<bind> lpoly_of c)
                        (\<lambda> f a poly. (Arctic_matrix ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (mat_ring (TYPE(arctic)) n)
        | Arctic_rat_mat n \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n MinInfty_delta CPF_Parser.arctic_rat_coeff) x \<bind> lpoly_of c)
                        (\<lambda>f a poly. (Arctic_rat_matrix ((f,a),(fst poly, fit_length c a (snd poly)))))))
                    (mat_ring (TYPE(rat arctic_delta)) n)
        )
      (\<lambda> type pi. case type of
            Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Int_carrier (map int_linear_poly pi) else Int_nl_carrier (map int_non_linear_poly pi)
          | Integer \<Rightarrow> Int_nl_carrier (map int_non_linear_poly pi)
          | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Rat_carrier (map rat_linear_poly pi) else Rat_nl_carrier d (map rat_non_linear_poly pi)
          | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Real_carrier (map real_linear_poly pi) else Real_nl_carrier d (map real_non_linear_poly pi)
          | Arctic \<Rightarrow> Arctic_carrier (map arctic_linear_poly pi)
          | Arctic_rat \<Rightarrow> Arctic_rat_carrier (map arctic_rat_linear_poly pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map int_matrix pi)
          | Rat_mat n sd \<Rightarrow> Rat_mat_carrier n sd (map rat_matrix pi)
          | Mini_Alg_mat n sd \<Rightarrow> Real_mat_carrier n sd (map real_matrix pi)
          | Arctic_mat n \<Rightarrow> Arctic_mat_carrier n (map arctic_matrix pi)
          | Arctic_rat_mat n \<Rightarrow> Arctic_rat_mat_carrier n (map arctic_rat_matrix pi)
      ))
    "

hide_const (open) "interpretation"

(**** adapted ****)

(* Just for easier xmlt to xmlt2 translation. For better maintanability prefer not to use them *)

definition xmlt2 :: "'a xmlt \<Rightarrow> 'a xmlt2" where
  "xmlt2 xmlt x \<equiv>
    case x of (xml,_,_,_,p) \<Rightarrow>
    (case xmlt xml of Right v \<Rightarrow> xml_return v ([],[],False,[],p) | Left err \<Rightarrow> xml_error err ([fst x],snd x))"

abbreviation xml_singleton :: "tag \<Rightarrow> 'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt2"
where "xml_singleton tag p f \<equiv> XMLdo tag {x \<leftarrow> p; xml_return (f x)}"

abbreviation xml_pair
where "xml_pair tag p1 p2 f \<equiv> XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; xml_return (f a b)}"

abbreviation xml_triple :: "tag \<Rightarrow> 'a xmlt2 \<Rightarrow> 'b xmlt2 \<Rightarrow> 'c xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt2"
where "xml_triple tag p1 p2 p3 f \<equiv> XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; xml_return (f a b c)}"

abbreviation xml_many where "xml_many tag p f \<equiv> XMLdo tag {a \<leftarrow>* p; xml_return (f a)}"
abbreviation xml_tuple4 where "xml_tuple4 tag p1 p2 p3 p4 f \<equiv>
  XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; d \<leftarrow> p4; xml_return (f a b c d)}"
abbreviation xml_tuple5 where "xml_tuple5 tag p1 p2 p3 p4 p5 f \<equiv>
  XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; d \<leftarrow> p4; e \<leftarrow> p5; xml_return (f a b c d e)}"



definition bounds_type :: "boundstype xmlt2" where
  "bounds_type \<equiv> xml_singleton ''type'' (
     xml_leaf ''roof'' Roof
     XMLor xml_leaf ''match'' Match
   ) id"

definition bounds_bound :: "nat xmlt2" where
  "bounds_bound \<equiv> xml_nat ''bound''"

definition state :: "string xmlt2" where
  "state = xml_text ''state''"

definition final_states :: "string list xmlt2" where
  "final_states = XMLdo ''finalStates'' {ret \<leftarrow>* state; xml_return ret}"

definition ta_bounds_lhs where
  "ta_bounds_lhs xml2name \<equiv> XMLdo ''lhs'' {
    XMLdo {
      a \<leftarrow> state;
      xml_return (Inr a)
    } XMLor XMLdo {
      f \<leftarrow> xml2name;
      h \<leftarrow> xml_nat ''height'';
      qs \<leftarrow>* state;
      xml_return (Inl ((f, h), qs))
    }
  }"

definition ta_normal_lhs where
  "ta_normal_lhs xml2name \<equiv> XMLdo ''lhs'' {
    XMLdo {
      a \<leftarrow> state;
      xml_return (Inr a)
    } XMLor XMLdo {
      a \<leftarrow> xml2name;
      b \<leftarrow>* state;
      xml_return (Inl (a,b))
    }
  }"

definition transition where
  "transition xml2lhs \<equiv> xml_pair ''transition''
    xml2lhs
    (xml_singleton ''rhs'' state id)
    (\<lambda>lhs q. case lhs of
      Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q)
    | Inr q' \<Rightarrow> Inr (q', q))"

definition transitions where
  "transitions xml2lhs = XMLdo ''transitions'' {
    rls \<leftarrow>* transition xml2lhs;
    let (rules, eps) = partition (\<lambda>rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
    let ruls = map (\<lambda>rl. case rl of Inl r \<Rightarrow> r) rules;
    let ep   = map (\<lambda>rl. case rl of Inr e \<Rightarrow> e) eps;
    xml_return (ruls, ep)
  }"

definition tree_automaton where
  "tree_automaton xml2lhs \<equiv> xml_pair ''treeAutomaton''
    final_states
    (transitions xml2lhs)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition closed_criterion :: "string ta_relation xmlt2" where
 "closed_criterion = XMLdo ''criterion'' {
    ret \<leftarrow>
      xml_leaf ''compatibility'' Id_Relation
      XMLor xml_singleton ''stateCompatibility'' (
        XMLdo ''relation'' {
          ret \<leftarrow>* xml_pair ''entry'' state state Pair;
          xml_return ret
        }
      ) Some_Relation
      XMLor xml_leaf ''decisionProcedure'' Decision_Proc
      XMLor xml_leaf ''decisionProcedureOld'' Decision_Proc_Old;
    xml_return ret
  }"

definition bounds_info :: "'a xmlt2 \<Rightarrow> ('a, string) bounds_info xmlt2" where
 "bounds_info xml2name \<equiv> XMLdo ''bounds'' {
    a \<leftarrow> bounds_type;
    b \<leftarrow> bounds_bound;
    c \<leftarrow> final_states;
    d \<leftarrow> tree_automaton (ta_bounds_lhs xml2name);
    e \<leftarrow>[Id_Relation] closed_criterion;
    xml_return (Bounds_Info a b c d e)
  }"

context
  fixes xml2name :: "'a :: show xmlt2"
begin

function ctxt ::  "('a, string) ctxt xmlt2"
where
 "ctxt x = (
    XMLdo ''box'' {xml_return Hole}
    XMLor XMLdo ''funContext'' {
      name \<leftarrow> xml2name;
      left \<leftarrow> XMLdo ''before'' {
        left \<leftarrow>* XTC_Parser.term xml2name;
        xml_return left
      };
      mid \<leftarrow> ctxt;
      right \<leftarrow> XMLdo ''after'' {
        right \<leftarrow>* XTC_Parser.term xml2name;
        xml_return right
      };
      xml_return (More name left mid right)
   }) x" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)


definition flat_contexts :: "('a, string) ctxt list xmlt2"
where "flat_contexts \<equiv>
  xml_do ''flatContexts'' (xml_take_many 0 \<infinity> (CPF_Parser.ctxt xml2name) xml_return)"

definition subst :: "('a, string) substL xmlt2"
where "subst \<equiv>
  XMLdo ''substitution'' {
    list \<leftarrow>* XMLdo ''substEntry'' {
      var \<leftarrow> xml_do ''var'' (xml_take_text xml_return);
      trm \<leftarrow> XTC_Parser.term xml2name;
      xml_return (var,trm)
    };
    xml_return list
  }
"


definition afs :: "'a afs_list xmlt2" where
  "afs \<equiv>
   XMLdo ''argumentFilter'' {
    ret \<leftarrow>* XMLdo ''argumentFilterEntry'' {
      name \<leftarrow> xml2name;
      arity \<leftarrow> xml_nat ''arity'';
      main \<leftarrow> xml_change (xml_nat ''collapsing'') (\<lambda>n. xml_return (Collapse (n - 1)))
        XMLor XMLdo ''nonCollapsing'' {
          ls \<leftarrow>* XTC_Parser.position;
          xml_return (AFList ls)
        };
      xml_return ((name, arity), main)
    };
    xml_return ret
  }"

definition proj :: "'a projL xmlt2" where
  "proj \<equiv> xml_change (CPF_Parser.afs xml2name) (\<lambda>afl. xml_return (Projection
    (map (\<lambda>(fa, e). case e of
      Collapse i \<Rightarrow> Pair fa i
    | _ \<Rightarrow> Pair fa (snd fa)) afl)))"


definition multiset_af :: "'a status_impl xmlt2" where
  "multiset_af \<equiv> XMLdo ''multisetArgumentFilter'' {
    ret \<leftarrow>* XMLdo ''multisetArgumentFilterEntry'' {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat ''arity'';
      p \<leftarrow> xml_do ''status'' (xml_take_many 0 \<infinity> XTC_Parser.position xml_return);
      xml_return ((f,a), p)
    };
    xml_return ret
  }"

definition
  relstep :: "(pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt2"
where
  "relstep \<equiv> XMLdo ''rewriteStep'' {
    p \<leftarrow> XTC_Parser.pos;
    r \<leftarrow> XTC_Parser.rule xml2name;
    rel \<leftarrow>? xml_leaf ''relative'' ();
    t \<leftarrow> XTC_Parser.term xml2name;
    xml_return (p, r, rel = None, t)
  }"

definition
  rstep :: "(pos \<times> ('a, string) rule \<times> ('a, string) term) xmlt2"
where
  "rstep = XMLdo ''rewriteStep'' {
    p \<leftarrow> XTC_Parser.pos;
    r \<leftarrow> XTC_Parser.rule xml2name;
    t \<leftarrow> XTC_Parser.term xml2name;
    xml_return (p, r, t)
  }"

definition
  estep :: "(pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt2"
where
  "estep \<equiv> XMLdo ''equationStep'' {
    p \<leftarrow> XTC_Parser.pos;
    r \<leftarrow> XTC_Parser.rule xml2name;
    b \<leftarrow> xml_leaf ''leftRight'' True XMLor xml_leaf ''rightLeft'' False;
    t \<leftarrow> XTC_Parser.term xml2name;
    xml_return (p, r, b, t)
  }"

definition
  relsteps :: "(('a,string)term \<times> ('a,string)prseq) xmlt2"
where
  "relsteps \<equiv> XMLdo ''rewriteSequence'' {
    start \<leftarrow> xml_singleton ''startTerm'' (XTC_Parser.term xml2name) id;
    steps \<leftarrow>* relstep;
    xml_return (start, steps)
  }"

definition
  rsteps :: "(('a,string)term \<times> ('a,string) rseq) xmlt2"
where
  "rsteps \<equiv> XMLdo ''rewriteSequence'' {
    start \<leftarrow> xml_singleton ''startTerm'' (XTC_Parser.term xml2name) id;
    steps \<leftarrow>* CPF_Parser.rstep xml2name;
    xml_return (start, steps)
  }"


definition pat_eqv_prf :: "('a,string)pat_eqv_prf xmlt2"
  where "pat_eqv_prf \<equiv> let sub = CPF_Parser.subst xml2name in
   xml_singleton ''patternEquivalence'' (
    XMLdo ''domainRenaming'' {n \<leftarrow> sub; xml_return (Pat_Dom_Renaming n)}
    XMLor XMLdo ''irrelevant'' {n \<leftarrow> sub; n2 \<leftarrow> sub; xml_return (Pat_Irrelevant n n2)}
    XMLor XMLdo ''simplification'' {n \<leftarrow> sub; n2 \<leftarrow> sub; xml_return (Pat_Simplify n n2)}
  ) id"

definition pat_term :: "(('a,string)term \<times> ('a,string)substL \<times> ('a,string)substL) xmlt2"
where "pat_term \<equiv> XMLdo ''patternTerm'' {
  t \<leftarrow> XTC_Parser.term xml2name;
  s1 \<leftarrow> CPF_Parser.subst xml2name;
  s2 \<leftarrow> CPF_Parser.subst xml2name;
  xml_return (t,s1,s2)
}"

abbreviation (input) pat_rule_fun where "pat_rule_fun x pat \<equiv> (let
   sub = CPF_Parser.subst xml2name;
   pt = pat_term;
   var = xml_text ''var''
in
(XMLdo ''patternRule'' {
  pt1 \<leftarrow> pt;
  pt2 \<leftarrow> pt;
  z \<leftarrow> XMLdo ''originalRule'' {
      r \<leftarrow> XTC_Parser.rule xml2name;
      b \<leftarrow> xml_bool ''isPair'';
      xml_return (Pat_OrigRule r b)
    } XMLor XMLdo ''initialPumping'' {
      a \<leftarrow> pat;
      b \<leftarrow> sub;
      c \<leftarrow> sub;
      xml_return (Pat_InitPump a b c)
    } XMLor XMLdo ''initialPumpingContext'' {
      a \<leftarrow> pat;
      b \<leftarrow> sub;
      c \<leftarrow> XTC_Parser.pos;
      d \<leftarrow> var;
      xml_return (Pat_InitPumpCtxt a b c d)
    } XMLor XMLdo ''equivalence'' {
      a \<leftarrow> pat;
      b \<leftarrow> xml_leaf ''left'' True XMLor xml_leaf ''right'' False;
      c \<leftarrow> CPF_Parser.pat_eqv_prf xml2name;
      xml_return (Pat_Equiv a b c)
    } XMLor XMLdo ''narrowing'' {
      a \<leftarrow> pat;
      b \<leftarrow> pat;
      c \<leftarrow> XTC_Parser.pos;
      xml_return (Pat_Narrow a b c)
    } XMLor XMLdo ''instantiation'' {
      a \<leftarrow> pat;
      b \<leftarrow> sub;
      c \<leftarrow> xml_leaf ''base'' Pat_Base XMLor xml_leaf ''pumping'' Pat_Pump XMLor xml_leaf ''closing'' Pat_Close;
      xml_return (Pat_Inst a b c)
    } XMLor XMLdo ''instantiationPumping'' {
      a \<leftarrow> pat;
      b \<leftarrow> xml_nat ''power'';
      xml_return (Pat_Exp_Sigma a b)
    } XMLor XMLdo ''rewriting'' {
      a \<leftarrow> pat;
      b \<leftarrow> CPF_Parser.rsteps xml2name;
      (po,va) \<leftarrow>
        xml_leaf ''base'' (Pat_Base,[])
        XMLor XMLdo ''pumping'' { v \<leftarrow> var; xml_return (Pat_Pump,v) }
        XMLor XMLdo ''closing'' { v \<leftarrow> var; xml_return (Pat_Close,v) };
      xml_return (Pat_Rewr a b po va)
    };
   xml_return z
  }) x)"

function pat_rule_prf :: "('a,string)pat_rule_prf xmlt2"
  where "pat_rule_prf x = pat_rule_fun x pat_rule_prf" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

(* one cannot put the 'let pat = ... ' directly into partial function, since then the monotonicity fails *)
lemma pat_rule_prf [code]:
  "pat_rule_prf x = (let pat = pat_rule_prf in pat_rule_fun x pat)"
     unfolding pat_rule_prf.simps[of x] Let_def by simp

definition
  conversion :: "(('a,string)term \<times> ('a,string) eseq) xmlt2"
where
  "conversion \<equiv> XMLdo ''conversion'' {
    a \<leftarrow> xml_singleton ''startTerm'' (XTC_Parser.term xml2name) id;
    b \<leftarrow>* CPF_Parser.estep xml2name;
    xml_return (a,b)
  }"

definition subsumption_proof :: "('a,string)subsumption_proof xmlt2"
where "subsumption_proof \<equiv> XMLdo ''subsumptionProof'' {
  ret \<leftarrow>* XMLdo ''ruleSubsumptionProof'' {
    r \<leftarrow> XTC_Parser.rule xml2name;
    (s,e) \<leftarrow> CPF_Parser.conversion xml2name;
    xml_return (r, e)
  };
  xml_return ret
}"

definition joinable_critical_pairs :: "('a,string)crit_pair_joins xmlt2"
where "joinable_critical_pairs \<equiv> let rew = CPF_Parser.rsteps xml2name in
  XMLdo ''joinableCriticalPairs'' {
    ret \<leftarrow>* XMLdo ''joinableCriticalPair'' {
      (s,sseq) \<leftarrow> rew;
      (t,tseq) \<leftarrow> rew;
      xml_return (s,sseq,t,tseq)
    };
  xml_return ret
}"

definition convertible_critical_peaks :: "('a,string) crit_peak_convs xmlt2"
where
  "convertible_critical_peaks \<equiv>
    let rew = CPF_Parser.rsteps xml2name in
    let conv = CPF_Parser.conversion xml2name in
    XMLdo ''convertibleCriticalPeaks'' {
      ret \<leftarrow>* XMLdo ''convertibleCriticalPeak'' {
        s \<leftarrow> xml_singleton ''source'' (XTC_Parser.term xml2name) id;
        (s1,cl1,sl,cl2) \<leftarrow> XMLdo ''conversionLeft'' {
          (s,c1) \<leftarrow> conv;
          (_,r) \<leftarrow> rew;
          (_,c2) \<leftarrow> conv;
          xml_return (s,c1,r,c2)
        };
        (s2,cr1,sr,cr2) \<leftarrow> XMLdo ''conversionRight'' {
          (s,c1) \<leftarrow> conv;
          (_,r) \<leftarrow> rew;
          (_,c2) \<leftarrow> conv;
          xml_return (s,c1,r,c2)
        };
        xml_return (s,s1,cl1,sl,cl2,s2,cr1,sr,cr2)
      };
      xml_return ret
  }"

definition wcr_proof :: "'a join_info xmlt2"
  where "wcr_proof \<equiv> xml_singleton ''wcrProof'' (
    xml_change (joinable_critical_pairs) (xml_return \<circ> Guided)
    XMLor xml_change (xml_nat ''joinableCriticalPairsBFS'') (xml_return \<circ> Join_BFS)
    XMLor xml_leaf ''joinableCriticalPairsAuto'' Join_NF
  ) id"

definition
  loop ::
    "(('a, string) term \<times> ('a, string) prseq \<times> ('a, string) substL \<times> ('a, string) ctxt) xmlt2"
where [code]:
  "loop \<equiv> XMLdo ''loop'' {
    (s,rseq) \<leftarrow> CPF_Parser.relsteps xml2name;
    \<sigma> \<leftarrow> CPF_Parser.subst xml2name;
    c \<leftarrow> CPF_Parser.ctxt xml2name;
    xml_return (s, rseq, \<sigma>, c)
  }"

definition
  start_term :: "('a, string) term \<Rightarrow> ('a, string) term xmlt2"
where
  "start_term t \<equiv> XMLdo ''startTerm'' {
    s \<leftarrow> XTC_Parser.term xml2name;
    if s = t then xml_return t
    else xml_error ''<startTerm> does not match lhs''
  }"

definition
  rseq ::
    "'a proj \<Rightarrow> ('a, string) rule \<Rightarrow> (('a,string)rule \<times> ('a,string)rseq)xmlt2"
where
  "rseq \<pi> r \<equiv> XMLdo ''rewriteSequence'' {
     _ \<leftarrow> CPF_Parser.start_term xml2name (proj_term \<pi> (fst r));
     rseq \<leftarrow>* CPF_Parser.rstep xml2name;
     xml_return (r, rseq)
  }"

definition
  projected_rseq ::
    "'a proj \<Rightarrow> (('a, string) rule \<times> ('a,string)rseq) xmlt2"
where
  "projected_rseq \<pi> \<equiv>
    XMLdo ''projectedRewriteSequence'' {
      r \<leftarrow> XTC_Parser.rule xml2name;
      ret \<leftarrow> CPF_Parser.rseq xml2name \<pi> r;
      xml_return ret
    }"


definition status_precedence :: "('a status_prec_repr) xmlt2" where
  "status_precedence \<equiv> XMLdo ''statusPrecedence'' {
    ret \<leftarrow>* XMLdo ''statusPrecedenceEntry'' {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat ''arity'';
      p \<leftarrow> xml_nat ''precedence'';
      s \<leftarrow> xml_leaf ''lex'' Lex XMLor xml_leaf ''mul'' Mul;
      xml_return ((f,a), (p,s))
    };
    xml_return ret
  }"

definition path_order :: "'a redtriple_impl xmlt2" where
  "path_order \<equiv> XMLdo ''pathOrder'' {
    prec\<tau> \<leftarrow> CPF_Parser.status_precedence xml2name;
    af \<leftarrow>[[]] CPF_Parser.afs xml2name;
    xml_return (RPO prec\<tau> af)
  }"

definition precedence_weight :: "(nat \<Rightarrow> 'a prec_weight_repr) xmlt2" where
  "precedence_weight \<equiv> XMLdo ''precedenceWeight'' {
    ret \<leftarrow>* XMLdo ''precedenceWeightEntry'' {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat ''arity'';
      p \<leftarrow> xml_nat ''precedence'';
      w \<leftarrow> xml_nat ''weight'';
      e \<leftarrow>? XMLdo ''subtermCoefficientEntries'' {
        ret \<leftarrow>* xml_nat ''entry'';
        xml_return ret
      };
      xml_return ((f,a), (p,w,e))
    };
    xml_return (Pair ret)
  }"

definition wpo_params :: "'a wpo_params xmlt2" where
  "wpo_params \<equiv> XMLdo ''precedenceStatus'' {
    ret \<leftarrow>* XMLdo ''precedenceStatusEntry'' {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat ''arity'';
      p \<leftarrow> xml_nat ''precedence'';
      s \<leftarrow> xml_do ''status'' (xml_take_many 0 \<infinity> XTC_Parser.position xml_return);
      xml_return ((f,a), (p,s))
    };
    xml_return ret
  }"

definition knuth_bendix_order :: "'a redtriple_impl xmlt2" where
  "knuth_bendix_order \<equiv> XMLdo ''knuthBendixOrder'' {
   w0 \<leftarrow> xml_nat ''w0'';
   prw \<leftarrow> CPF_Parser.precedence_weight xml2name;
   afs \<leftarrow>[[]] CPF_Parser.afs xml2name;
   xml_return (KBO (prw w0) afs)
  }"

function redtriple :: "bool \<Rightarrow> 'a redtriple_impl xmlt2" where
  "redtriple bi x = (
    CPF_Parser.path_order xml2name
    XMLor CPF_Parser.knuth_bendix_order xml2name
    XMLor XMLdo ''weightedPathOrder'' {
      a \<leftarrow> CPF_Parser.wpo_params xml2name;
      b \<leftarrow> redtriple bi;
      xml_return (WPO a b)
    }
    XMLor xmlt2 (CPF_Parser.interpretation bi (xmlt_of_xmlt2 xml2name))
   ) x" by (pat_completeness, auto)

termination by (relation "measure (size_xml_state \<circ> snd)", auto)

definition level_mapping :: "'a scnp_af xmlt2" where
  "level_mapping = XMLdo ''levelMapping'' {
    ret \<leftarrow>* XMLdo ''levelMappingEntry'' {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat ''arity'';
      ps \<leftarrow>* xml_pair ''positionLevelEntry''
         (xml_nat ''position'')
         (xml_nat ''level'')
         Pair;
      xml_return ((f,a), map (\<lambda> (p,l). (if p = 0 then a else p - 1,l)) ps)
    };
    xml_return ret
  }"

definition redtriple' where
  "redtriple' bi \<equiv> XMLdo ''redPair'' {
    ret \<leftarrow> XMLdo ''scnp'' {
      a \<leftarrow> XMLdo ''status'' {
        ret \<leftarrow>
          xml_leaf ''ms'' MS_Ext
          XMLor xml_leaf ''min'' Min_Ext
          XMLor xml_leaf ''dms'' Dms_Ext
          XMLor xml_leaf ''max'' Max_Ext;
        xml_return ret
      };
      b \<leftarrow> level_mapping;
      c \<leftarrow> xml_singleton ''redPair'' (redtriple False) id;
      xml_return (Inl (SCNP a b c))
    } XMLor xml_change (redtriple bi) (xml_return \<circ> Inr);
  xml_return ret
}"


definition
  ordering_constraint_proof :: "bool \<Rightarrow> 'a redtriple_impl xmlt2"
where
  "ordering_constraint_proof bi \<equiv>
    xml_singleton ''orderingConstraintProof'' (
      xml_singleton ''redPair'' (redtriple bi) id
    ) id"

definition
  ordering_constraint_proof'
where
  "ordering_constraint_proof' bi \<equiv>
     xml_singleton ''orderingConstraintProof'' (redtriple' bi) id"

definition
  id_labeling :: "(('f,'a list)lab,'a,'a list + ('f,'a list)lab list)label"
where
  "id_labeling _ cs = Inl cs"

definition scg_position :: "nat xmlt2" where
  "scg_position \<equiv> xml_nat ''position''"

definition
  scg :: "(('a, string) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) xmlt2"
where
  "scg \<equiv> XMLdo ''sizeChangeGraph'' {
    lr \<leftarrow> XTC_Parser.rule xml2name;
    edges \<leftarrow>* XMLdo ''edge'' {
       p \<leftarrow> CPF_Parser.scg_position;
       s \<leftarrow> xml_bool ''strict'';
       q \<leftarrow> CPF_Parser.scg_position;
       xml_return (s,(p,q))
    };
    xml_return (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges)))
  }"

definition uncurry_info :: "('a, string) uncurry_info xmlt2" where
  "uncurry_info \<equiv> XMLdo ''uncurryInformation'' {
    a \<leftarrow> xml2name;
    sml \<leftarrow> XMLdo ''uncurriedSymbols'' {
      ret \<leftarrow>* XMLdo ''uncurriedSymbolEntry'' {
        f \<leftarrow> xml2name;
        n \<leftarrow> xml_nat ''arity'';
        fs \<leftarrow>* xml2name;
        xml_return ((f, n), fs)
      };
      xml_return ret
    };
    U \<leftarrow> xml_singleton ''uncurryRules'' (XTC_Parser.rules xml2name) id;
    E \<leftarrow> xml_singleton ''etaRules'' (XTC_Parser.rules xml2name) id;
    xml_return (a, sml, U, E)
  }"


definition xml2dp_input
where
  "xml2dp_input \<equiv> XMLdo ''dpInput'' {
    r \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
    p \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
    s \<leftarrow>[No_Strategy] strategy xml2name;
    m \<leftarrow> xml_bool ''minimal'';
    xml_return (DP_input m p s r)
  }"

function
  xml2cond_constraint :: "('a,string) cond_constraint xmlt2" where
  "xml2cond_constraint x = xml_singleton ''conditionalConstraint'' (
    XMLdo ''all'' {
      a \<leftarrow> xml_text ''var'';
      b \<leftarrow> xml2cond_constraint;
      xml_return (CC_all a b)
    } XMLor XMLdo ''implication'' {
      c \<leftarrow> xml2cond_constraint;
      cs \<leftarrow>* xml2cond_constraint;
      let ccs = c # cs;
      xml_return (CC_impl (take (length cs) ccs) (last ccs))
    } XMLor XMLdo ''constraint'' {
      s \<leftarrow> XTC_Parser.term xml2name;
      rel \<leftarrow>
        xml_leaf ''rewrite'' None
        XMLor xml_leaf ''strict'' (Some True)
        XMLor xml_leaf ''nonStrict'' (Some False);
      t \<leftarrow> XTC_Parser.term xml2name;
      xml_return (case rel of None \<Rightarrow> CC_rewr s t | Some stri \<Rightarrow> CC_cond stri (s,t))
    }) id x" by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)


function xml2cond_constraint_prf :: "('a,string) cond_constraint_prf xmlt2" where
  "xml2cond_constraint_prf x = (
  let cc = xml2cond_constraint in
  xml_singleton ''conditionalConstraintProof'' (
    xml_leaf ''final'' Final
    XMLor XMLdo ''differentConstructor'' {
      a \<leftarrow> cc;
      xml_return (Different_Constructor a)
    } XMLor XMLdo ''sameConstructor'' {
      a \<leftarrow> cc; b \<leftarrow> cc; c \<leftarrow> xml2cond_constraint_prf;
      xml_return (Same_Constructor a b c)
    } XMLor XMLdo ''variableEquation'' {
      a \<leftarrow> xml_text ''var''; b \<leftarrow> XTC_Parser.term xml2name; c \<leftarrow>  cc;
      d \<leftarrow> xml2cond_constraint_prf;
      xml_return (Variable_Equation a b c d)
    } XMLor XMLdo ''funargIntoVar'' {
      a \<leftarrow> cc;
      b \<leftarrow> XTC_Parser.position;
      c \<leftarrow> xml_text ''var'';
      d \<leftarrow> cc;
      e \<leftarrow> (xml2cond_constraint_prf);
      xml_return (Funarg_Into_Var a b c d e)
    } XMLor XMLdo ''simplifyCondition'' {
      a \<leftarrow> cc; b \<leftarrow> (CPF_Parser.subst xml2name); c \<leftarrow> cc; d \<leftarrow> (xml2cond_constraint_prf);
      xml_return (Simplify_Condition a b c d)
    } XMLor XMLdo ''induction'' {
      a \<leftarrow> cc;
      b \<leftarrow> XMLdo ''conjuncts'' {ret \<leftarrow>* cc; xml_return ret};
      c \<leftarrow> XMLdo ''ruleConstraintProofs'' {
        ret \<leftarrow>* XMLdo ''ruleConstraintProof'' {
          lr \<leftarrow> XTC_Parser.rule xml2name;
          rys \<leftarrow> XMLdo ''subtermVarEntries'' {
            ret \<leftarrow>* XMLdo ''subtermVarEntry'' {
              a \<leftarrow> XTC_Parser.term xml2name;
              b \<leftarrow>* xml_text ''var'';
              xml_return (a,b)
            };
            xml_return ret
          };
          cc \<leftarrow> cc;
          p \<leftarrow> xml2cond_constraint_prf;
          xml_return (lr,rys,cc,p)
        };
        xml_return ret
      };
      xml_return (Induction a b c)
    } XMLor XMLdo ''deleteCondition'' {
      a \<leftarrow> cc;
      b \<leftarrow> xml2cond_constraint_prf;
      xml_return (Delete_Condition a b)
   }) id
    x)"
by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

definition
  xml2cond_red_pair_proof :: "('a, string) cond_red_pair_prf xmlt2" where
  "xml2cond_red_pair_proof \<equiv> XMLdo ''condRedPairProof'' {
    c \<leftarrow> xml2name;
    b \<leftarrow> xml_nat ''before'';
    a \<leftarrow> xml_nat ''after'';
    ccs \<leftarrow> XMLdo ''conditions'' {
      ret \<leftarrow>* XMLdo ''condition'' {
         c \<leftarrow> xml2cond_constraint;
         s \<leftarrow> xml_singleton ''dpSequence'' (XTC_Parser.rules xml2name) id;
         p \<leftarrow> xml2cond_constraint_prf;
         xml_return (c,s,p)
      };
      xml_return ret
    };
    xml_return (Cond_Red_Pair_Prf c ccs b a)
  }"

definition renaming :: "('a \<times> 'a)list xmlt2" where
  "renaming  \<equiv> XMLdo ''renaming'' {
    ret \<leftarrow>* XMLdo ''renamingEntry'' {
      a \<leftarrow> xml2name;
      b \<leftarrow> xml2name;
      xml_return (a,b)
    };
    xml_return ret
  }"

(* TODO: improve CPF and remove this *)
definition xml2complexity_input
where "xml2complexity_input \<equiv> XMLdo ''complexityInput'' {
    (q, r, s, start) \<leftarrow> xml_change (xml2_trs_input xml2name) (\<lambda>inp.
      case inp of Inn_TRS_input q r s start \<Rightarrow> xml_return (q, r, s, start)
      |  _ \<Rightarrow> xml_error ''require innermost trs at this point''
    );
    cm \<leftarrow> complexity_measure xml2name;
    cc \<leftarrow> complexity_class;
    xml_return (CPX_input q r s cm cc)
  }"

definition symbols_no_arity :: "string \<Rightarrow> 'a list xmlt2"
where
  "symbols_no_arity tagname =
    XMLdo tagname {ret \<leftarrow>* xml2name; xml_return ret}"

definition
  xml2ac_tp_input :: "('a,string) input xmlt2"
where
  "xml2ac_tp_input \<equiv> XMLdo ''acRewriteSystem'' {
     a \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
     b \<leftarrow> symbols_no_arity ''Asymbols'';
     c \<leftarrow> symbols_no_arity ''Csymbols'';
     xml_return (AC_input a b c)
  }"

definition
  xml2input :: "('a,string) input xmlt2"
where
  "xml2input \<equiv> xml_singleton ''input'' (
    xml2dp_input
    XMLor xml2completion_input xml2name
    XMLor xml2ordered_completion_input xml2name
    XMLor xml2_trs_input xml2name
    XMLor xml2equational_input xml2name
    XMLor xml2complexity_input
    XMLor xml2ctrs_input xml2name
    XMLor XMLdo ''treeAutomatonProblem'' {
      a \<leftarrow> tree_automaton (ta_normal_lhs xml2name);
      b \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      xml_return (TA_input a b)
    } XMLor xml2ac_tp_input
    XMLor xml_change IA.lts_input_parser (xml_return \<circ> LTS_input)
    XMLor xml_change IA.lts_safety_input_parser (\<lambda>(x,y). xml_return (LTS_safety_input x y))
    XMLor xml_change xml2unknown_input (xml_return \<circ> Unknown_input)
  ) id"


text \<open>
  Parse a list of pairs of rules, where the first argument gives the name of the outermost tag
  and the second argument the name of the tag around each pair.
\<close>
definition
  rule_pairs :: "string \<Rightarrow> string \<Rightarrow> (('a, string) rule \<times> ('a, string) rule) list xmlt2"
where
  "rule_pairs s p \<equiv> XMLdo s {
    ret \<leftarrow>* xml_pair p
      (XTC_Parser.rule xml2name)
      (XTC_Parser.rule xml2name)
      Pair;
    xml_return ret
  }"

definition xml2inn_fp_trs_assm where
  "xml2inn_fp_trs_assm \<equiv> xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of
      Inn_TRS_input inn r s start \<Rightarrow>
        if start \<noteq> start_term.Full then
          xml_error ''start term is not allowed here''
        else
          xml_return (Inl (default_nfs_trs, strategy_to_Q inn r, r, s))
    | FP_TRS_input fp r \<Rightarrow> xml_return (Inr (strategy_to_fp fp r, r))
    | _ \<Rightarrow> xml_error ''trs as input required''
  )"

definition xml2inn_trs_assm where
  "xml2inn_trs_assm \<equiv> xml_change xml2inn_fp_trs_assm (\<lambda>inp.
    case inp of
      Inl i \<Rightarrow> xml_return i
    | _ \<Rightarrow> xml_error ''innermost (relative) TRS expected at this point''
   )"

definition xml2fp_trs_assm where
  "xml2fp_trs_assm \<equiv> xml_change xml2inn_fp_trs_assm (\<lambda>inp.
    case inp of
      Inr i \<Rightarrow> xml_return i
    | _ \<Rightarrow> xml_error ''FP TRS expected at this point''
   )"

definition xml2dp_input' where
  "xml2dp_input' termination \<equiv> xml_change xml2dp_input
    (\<lambda>ret. case ret of DP_input m p q r \<Rightarrow>
     xml_return (if termination then default_nfs_dp else default_nfs_nt_dp, m, p, [], strategy_to_Q q r, [], r))"

definition xml2complexity_input' where
  "xml2complexity_input' \<equiv> xml_change xml2complexity_input
    (\<lambda> ret. case ret of CPX_input q s w cm cc \<Rightarrow> xml_return (strategy_to_Q q (s @ w), s, w, cm,cc))"
end

context fixes xml2name :: "('f :: {show,compare_order}, 'l :: {show,compare_order}) lab xmlt2"
begin
function
  xml2dp_termination_proof :: "('f, 'l, string) dp_termination_proof xmlt2"
  and xml2trs_termination_proof :: "('f, 'l, string) trs_termination_proof xmlt2"
  and xml2trs_termination_proof_inner
  and xml2unknown_proof :: "('f, 'l, string) unknown_proof xmlt2" where
  "xml2dp_termination_proof x = (
    xml_singleton ''dpProof'' (
    xml_leaf ''pIsEmpty'' P_is_Empty
    XMLor xml_many ''depGraphProc'' (
      XMLdo ''component'' {
        dps \<leftarrow> (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id);
        _ \<leftarrow>? xml_bool ''realScc''; (* this is redundant, and the original syntax is not expressible by XML scheme *)
        _ \<leftarrow>? XMLdo ''arcs'' { _ \<leftarrow>* xml_any; xml_return () };
        prfOpt \<leftarrow>? xml2dp_termination_proof;
        xml_return (prfOpt, dps)
      }
    ) Dep_Graph_Proc
    XMLor xml_triple ''redPairProc''
      (CPF_Parser.ordering_constraint_proof' xml2name False)
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Redpair_Proc
    XMLor xml_pair ''usableRulesProc''
      (xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Usable_Rules_Proc
   XMLor xml_pair ''innermostLhssRemovalProc''
      (innermostLhss xml2name)
      xml2dp_termination_proof
      Q_Reduction_Proc
    XMLor XMLdo ''rewritingProc'' {
      (s,t) \<leftarrow> XTC_Parser.rule xml2name;
      (p,lr,t') \<leftarrow> CPF_Parser.rstep xml2name;
      st'' \<leftarrow>[(s,t')] XTC_Parser.rule xml2name;
      ur \<leftarrow>? xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id;
      prof \<leftarrow> xml2dp_termination_proof;
      xml_return (Rewriting_Proc ur (s,t) (s,t') st'' lr p prof)
    }
    XMLor xml_tuple4 ''narrowingProc''
      (XTC_Parser.rule xml2name)
      XTC_Parser.pos
      (xml_singleton ''narrowings'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Narrowing_Proc
    XMLor xml_triple ''instantiationProc''
      (XTC_Parser.rule xml2name)
      (xml_singleton ''instantiations'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Instantiation_Proc
    XMLor XMLdo ''forwardInstantiationProc'' {
      a \<leftarrow> XTC_Parser.rule xml2name;
      b \<leftarrow> xml_singleton ''instantiations'' (XTC_Parser.rules xml2name) id;
      c \<leftarrow>? xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml2dp_termination_proof;
      xml_return (Forward_Instantiation_Proc a b c d)
    } XMLor XMLdo ''semlabProc'' {
      sli \<leftarrow> xmlt2 (CPF_Parser.sl_variant (xmlt_of_xmlt2 xml2name));
      lp \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
      lr \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      lq \<leftarrow>[[]] innermostLhss xml2name;
      p \<leftarrow> xml2dp_termination_proof;
      xml_return (Semlab_Proc sli lp lq lr p)
    } XMLor XMLdo ''subtermProc'' {
      pi_mpi \<leftarrow> xml_change (CPF_Parser.proj xml2name) (xml_return \<circ> Inl)
        XMLor xml_change (CPF_Parser.multiset_af xml2name) (xml_return \<circ> Inr);
      seq \<leftarrow>* CPF_Parser.projected_rseq xml2name (create_proj (case pi_mpi of Inl lpi \<Rightarrow> lpi | Inr mpi \<Rightarrow> Projection []));
      dps \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_termination_proof;
      xml_return (case pi_mpi of
         Inl lpi \<Rightarrow> Subterm_Criterion_Proc lpi seq dps prf
       | Inr mpi \<Rightarrow> Gen_Subterm_Criterion_Proc mpi dps prf)
    } XMLor xml_tuple4 ''redPairUrProc''
      (CPF_Parser.ordering_constraint_proof' xml2name False)
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      (xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Redpair_UR_Proc
    XMLor xml_tuple5 ''monoRedPairUrProc''
      (CPF_Parser.ordering_constraint_proof xml2name False)
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      (xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Mono_Redpair_UR_Proc
    XMLor xml_tuple4 ''monoRedPairProc''
      (CPF_Parser.ordering_constraint_proof xml2name False)
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Mono_Redpair_Proc
    XMLor xml_triple ''innermostMonoRedPairProc''
      (CPF_Parser.ordering_constraint_proof xml2name False)
      (xml_pair ''deleted''
        (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
        (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
        Pair)
      xml2dp_termination_proof
      (\<lambda> rp (p,r). Mono_URM_Redpair_Proc rp p r)
    XMLor XMLdo ''uncurryProc'' {
      a \<leftarrow>? xml_nat ''applicativeTop'';
      b \<leftarrow> CPF_Parser.uncurry_info xml2name;
      c \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      e \<leftarrow> xml2dp_termination_proof;
      xml_return (Uncurry_Proc a b c d e)
    } XMLor XMLdo ''flatContextClosureProc'' {
      a \<leftarrow> xml_singleton ''freshSymbol'' xml2name id;
      b \<leftarrow> CPF_Parser.flat_contexts xml2name;
      c \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      e \<leftarrow> xml2dp_termination_proof;
      xml_return (Fcc_Proc a b c d e)
    } XMLor xml_pair ''switchInnermostProc''
      (CPF_Parser.wcr_proof xml2name)
      xml2dp_termination_proof
      Switch_Innermost_Proc
    XMLor xml_tuple4 ''splitProc''
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2dp_termination_proof
      xml2dp_termination_proof
      Split_Proc
    XMLor xml_singleton ''finitenessAssumption''
      (xml2dp_input' xml2name True)
      (\<lambda> dpp. Assume_Finite dpp [])
    XMLor XMLdo ''unknownProof'' {
      a \<leftarrow> xml_text ''description'';
      b \<leftarrow> xml2dp_input' xml2name True;
      c \<leftarrow>* xml_any; (* TODO *)
      xml_return (Assume_Finite b [])
    } XMLor xml_singleton ''switchToTRS''
         xml2trs_termination_proof
         To_Trs_Proc
    XMLor xml_triple ''unlabProc''
         (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
         (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
         xml2dp_termination_proof
         Unlab_Proc
    XMLor XMLdo ''generalRedPairProc'' {
      rp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name True;
      s \<leftarrow> xml_singleton ''strict'' (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml_singleton ''bound'' (XTC_Parser.rules xml2name) id;
      c \<leftarrow> CPF_Parser.xml2cond_red_pair_proof xml2name;
      ps \<leftarrow> xml2dp_termination_proof;
      pbo \<leftarrow>? xml2dp_termination_proof;
      xml_return (General_Redpair_Proc rp s b c (case pbo of None \<Rightarrow> [ps] | Some pb \<Rightarrow> [ps,pb]))
    } XMLor xml_triple ''complexConstantRemovalProc''
         (XTC_Parser.term xml2name)
         (xml_many ''ruleMap''
            (xml_pair ''ruleMapEntry''
              (XTC_Parser.rule xml2name)
              (XTC_Parser.rule xml2name)
              Pair) id)
         xml2dp_termination_proof
         (\<lambda> t rls prf. Complex_Constant_Removal_Proc (Complex_Constant_Removal_Proof t rls) prf)
    XMLor XMLdo ''sizeChangeProc'' {
      version \<leftarrow> xml_leaf ''subtermCriterion'' None
        XMLor XMLdo ''reductionPair'' {
          redp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
          ur \<leftarrow>? xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id;
          xml_return (Some (redp,ur))
        };
      b \<leftarrow>* CPF_Parser.scg xml2name;
      case version of
          None \<Rightarrow> xml_return (Size_Change_Subterm_Proc b)
        | Some redp_ur \<Rightarrow> xml_return (Size_Change_Redpair_Proc (fst redp_ur) (snd redp_ur) b)
    }
  ) id) x"
| "xml2trs_termination_proof_inner x = (
    xml_leaf ''rIsEmpty'' R_is_Empty
    XMLor XMLdo ''sIsEmpty'' {
      p \<leftarrow> xml2trs_termination_proof;
      xml_return p
    } XMLor XMLdo ''semlab'' {
       sli \<leftarrow> xmlt2 (CPF_Parser.sl_variant (xmlt_of_xmlt2 xml2name));
       lr \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
       lrw \<leftarrow>[[]] xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
       lq \<leftarrow>[[]] innermostLhss xml2name;
       p \<leftarrow> xml2trs_termination_proof;
       xml_return (Semlab sli lq (lr@lrw) p)
    } XMLor xml_triple ''split''
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2trs_termination_proof
      xml2trs_termination_proof
      Split
    XMLor xml_triple ''dpTrans''
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      (xml_bool ''markedSymbols'')
      xml2dp_termination_proof
      (\<lambda>dps _ prf. DP_Trans default_nfs_dp True dps prf)
    XMLor XMLdo ''ruleRemoval'' {
      ord \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      r \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      s \<leftarrow>[[]] xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_termination_proof;
      xml_return (Rule_Removal ord (r@s) p)
    }
    XMLor xml_change (bounds_info xml2name) (xml_return \<circ> Bounds)
    XMLor XMLdo ''stringReversal'' {
      _ \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      _ \<leftarrow>? xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2trs_termination_proof;
      xml_return (String_Reversal prf)
    }
    XMLor xml_singleton ''equalityRemoval''
       xml2trs_termination_proof
       Drop_Equality
    XMLor XMLdo ''constantToUnary'' {
       v \<leftarrow> XTC_Parser.plain_var;
       ren \<leftarrow> CPF_Parser.renaming xml2name;
       S \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
       Sw \<leftarrow>[[]] xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
       p \<leftarrow> xml2trs_termination_proof;
       xml_return (Constant_String (Const_string_sound_proof v ren S Sw) p)
    } XMLor xml_pair ''removeNonApplicableRules''
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2trs_termination_proof
      Remove_Nonapplicable_Rules
    XMLor XMLdo ''uncurry'' {
      i \<leftarrow> CPF_Parser.uncurry_info xml2name;
      r \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      s \<leftarrow>[[]] xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_termination_proof;
      xml_return (Uncurry i (r@s) p)
    } XMLor XMLdo ''flatContextClosure'' {
      i \<leftarrow> CPF_Parser.flat_contexts xml2name;
      r \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      s \<leftarrow>[[]] xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_termination_proof;
      xml_return (Fcc i (r@s) p)
    } XMLor xml_pair ''switchInnermost''
      (CPF_Parser.wcr_proof xml2name)
      xml2trs_termination_proof
      Switch_Innermost
    XMLor xml_pair ''permutingArgumentFilter''
      (CPF_Parser.afs xml2name)
      xml2trs_termination_proof
      Permuting_AFS
    XMLor xml_singleton ''terminationAssumption''
        (xml2inn_trs_assm xml2name)
        (\<lambda> qtrs. Assume_SN qtrs [])
    XMLor XMLdo ''relativeTerminationAssumption'' {
       ret \<leftarrow> XMLdo ''trsInput'' {
         r \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
         rw \<leftarrow>[[]] xml_singleton ''relativeRules'' (XTC_Parser.rules xml2name) id;
         xml_return (Assume_SN (default_nfs_trs, [], r, rw) [])
      };
      xml_return ret
    } XMLor XMLdo ''unknownProof'' {
        a \<leftarrow> xml_text ''description'';
        b \<leftarrow> xml2inn_trs_assm xml2name;
        c \<leftarrow>* xml_any;
        xml_return (Assume_SN b [])
    }
  ) x"
| "xml2trs_termination_proof x = (
    xml_singleton ''trsTerminationProof'' xml2trs_termination_proof_inner id
    XMLor xml_singleton ''relativeTerminationProof'' xml2trs_termination_proof_inner id
  ) x"
| "xml2unknown_proof x = (XMLdo ''unknownInputProof'' {
    ret \<leftarrow>
      xml_singleton ''unknownAssumption''
        xml2unknown_input
        (\<lambda> u. Assume_Unknown u [])
      XMLor XMLdo ''unknownProof'' {
        desc \<leftarrow> xml_text ''description'';
        b \<leftarrow> xml2unknown_input;
        c \<leftarrow>* xml_any;
        xml_return (Assume_Unknown b [])
      };
    xml_return ret
  }) x"
by (pat_completeness, auto)


lemma drop_size_xml[dest]:
  assumes y: "drop n z = y" shows "sum_list (map size_xml y) \<le> sum_list (map size_xml z)"
proof (fold y, induct z arbitrary:n y)
  case Nil
  then show ?case by simp
next
  case (Cons a z)
  from le_trans[OF this le_add2] show ?case by (cases n, auto)
qed

termination
by (relation "measure (\<lambda>x. let y = case x of Inl (Inl y) \<Rightarrow> y | Inl (Inr y) \<Rightarrow> y | Inr (Inl y) \<Rightarrow> y | Inr (Inr y) \<Rightarrow> y in size_xml_state y)", auto simp: Let_def split:sum.split)


function
  xml2complexity_proof :: "('f, 'l, string) complexity_proof xmlt2"
where
  [code]: "xml2complexity_proof x = xml_singleton ''complexityProof'' (
    XMLdo ''ruleShifting'' {
      rp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name True;
      del \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      ur \<leftarrow>? xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2complexity_proof;
      xml_return (Rule_Shift_Complexity rp del ur prf)
    } XMLor xml_pair ''usableRules''
      (xml_singleton ''nonUsableRules'' (XTC_Parser.rules xml2name) id)
      xml2complexity_proof
      Usable_Rules_Complexity
    XMLor xml_triple ''split''
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2complexity_proof
      xml2complexity_proof
      Split_Complexity
    XMLor xml_pair ''removeNonApplicableRules''
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2complexity_proof
      Remove_Nonapplicable_Rules_Complexity
    XMLor xml_change (bounds_info xml2name) (xml_return \<circ> Matchbounds_Complexity)
    XMLor xml_triple ''relativeBounds''
         (bounds_info xml2name)
         (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
         xml2complexity_proof
         Matchbounds_Rel_Complexity
    XMLor xml_leaf ''rIsEmpty''
      RisEmpty_Complexity
    XMLor xml_tuple4 ''dtTransformation''
        (CPF_Parser.rule_pairs xml2name ''strictDTs'' ''ruleWithDT'')
        (CPF_Parser.rule_pairs xml2name ''weakDTs'' ''ruleWithDT'')
        (innermostLhss xml2name)
        xml2complexity_proof
        (\<lambda> S W inn p. DT_Transformation (DT_Transformation_Info S W inn) p)
    XMLor xml_tuple5 ''wdpTransformation''
      (XTC_Parser.symbols xml2name ''compoundSymbols'')
      (CPF_Parser.rule_pairs  xml2name ''strictWDPs'' ''ruleWithWDP'')
      (CPF_Parser.rule_pairs  xml2name ''weakWDPs'' ''ruleWithWDP'')
      (innermostLhss xml2name)
      xml2complexity_proof
      (\<lambda>Comp S W Q' p. WDP_Transformation (WDP_Trans_Info (set Comp) S W Q') p)
    XMLor XMLdo ''unknownProof'' {
      a \<leftarrow> xml_text ''description'';
      b \<leftarrow> xml2complexity_input' xml2name;
      c \<leftarrow>* xml_pair ''subProof''
      (xml2complexity_input' xml2name)
      xml2complexity_proof
      Complexity_assm_proof;
      xml_return (Complexity_Assumption b c)
    } XMLor xml_singleton ''complexityAssumption''
       (xml2complexity_input' xml2name)
       (\<lambda> p. Complexity_Assumption p [])
  ) id x"
by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)


function xml2ac_dp_termination_proof :: "(_, string) ac_dp_termination_proof xmlt2"
where "xml2ac_dp_termination_proof x = xml_singleton ''acDPTerminationProof'' (
    xml_tuple4 ''acRedPairProc''
          (CPF_Parser.ordering_constraint_proof xml2name False)
          (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
          (xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id)
          xml2ac_dp_termination_proof
          AC_Redpair_UR_Proc
    XMLor xml_triple ''acSubtermProc''
      (CPF_Parser.multiset_af xml2name)
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      xml2ac_dp_termination_proof
      (\<lambda> pi r prf. AC_Subterm_Proc pi r prf)
    XMLor xml_leaf ''acTrivialProc'' AC_P_is_Empty
    XMLor xml_tuple5 ''acMonoRedPairProc''
          (CPF_Parser.ordering_constraint_proof xml2name False)
          (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
          (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
          (xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id)
          xml2ac_dp_termination_proof
          AC_Mono_Redpair_UR_Proc
    XMLor XMLdo ''acDepGraphProc'' {
      ret \<leftarrow>* XMLdo ''component'' {
        dps \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
        _ \<leftarrow>? xml_bool ''realScc''; (* redundant *)
        prfOpt \<leftarrow>? xml2ac_dp_termination_proof;
        xml_return (prfOpt, dps)
      };
      xml_return (AC_Dep_Graph_Proc ret)
    }
   ) id x"
by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

function xml2ac_termination_proof :: "('f, 'l, string) ac_termination_proof xmlt2"
where "xml2ac_termination_proof x = xml_singleton ''acTerminationProof'' (
    XMLdo ''acDependencyPairs'' {
      e \<leftarrow> xml_singleton ''equations'' (XTC_Parser.rules xml2name) id;
      dpe \<leftarrow> xml_singleton ''dpEquations'' (XTC_Parser.rules xml2name) id;
      dp \<leftarrow> xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
      rext \<leftarrow> xml_singleton ''extensions'' (XTC_Parser.rules xml2name) id;
      p1 \<leftarrow> xml2ac_dp_termination_proof;
      p2opt \<leftarrow>? xml2ac_dp_termination_proof;
      case p2opt
      of Some p2 \<Rightarrow> xml_return (AC_DP_Trans (AC_dependency_pairs_proof e dp dpe rext) p1 p2)
      | None \<Rightarrow> xml_return (AC_DP_Trans_Single (AC_dependency_pairs_proof e dp dpe rext) p1)
    } XMLor xml_triple ''acRuleRemoval''
      (CPF_Parser.ordering_constraint_proof xml2name False)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2ac_termination_proof
      AC_Rule_Removal
    XMLor  xml_leaf ''acRIsEmpty''
      AC_R_is_Empty
   ) id x"
by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

definition nonloop :: "(_,string) non_loop_prf xmlt2" where
  "nonloop \<equiv> XMLdo ''nonLoop'' {
    a \<leftarrow> CPF_Parser.pat_rule_prf xml2name;
    b \<leftarrow> CPF_Parser.subst xml2name;
    c \<leftarrow> CPF_Parser.subst xml2name;
    d \<leftarrow> xml_nat ''natural'';
    e \<leftarrow> xml_nat ''natural'';
    f \<leftarrow> XTC_Parser.pos;
    xml_return (Non_loop_prf a b c d e f)
  }"

definition string :: "_ list xmlt2" where
  "string \<equiv> xml_many ''string'' xml2name id"

definition oc_srs :: "_ srs_rule xmlt2" where
  "oc_srs = xml_pair ''overlapClosureSRS'' string string Pair"

definition word_pattern :: "_ word_pat xmlt2" where
  "word_pattern \<equiv>
    XMLdo ''wordPattern'' {
      l \<leftarrow> string;
      m \<leftarrow> string;
      f \<leftarrow> xml_nat ''factor'';
      c \<leftarrow> xml_nat ''constant'';
      r \<leftarrow> string;
      xml_return (l,(f,c,m),r)
  }"

definition derivation_pattern :: "_ deriv_pat xmlt2" where
  "derivation_pattern \<equiv> xml_pair ''derivationPattern''
     word_pattern word_pattern Pair"

abbreviation "xml_tuple6 tag p1 p2 p3 p4 p5 p6 t \<equiv>
 XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; d \<leftarrow> p4; e \<leftarrow> p5; f \<leftarrow> p6; xml_return (t a b c d e f)}"

definition derivation_pattern_proof :: "_ dp_proof_step xmlt2" where
  "derivation_pattern_proof \<equiv>
    (let
      oc = oc_srs;
      dp = derivation_pattern;
      s = string in
    xml_singleton ''derivationPatternProof'' (
      xml_pair ''OC1'' oc (xml_bool ''isPair'') OC1
    XMLor xml_tuple6 ''OC2'' oc oc oc s s s OC2
    XMLor xml_tuple6 ''OC2prime'' oc oc oc s s s OC2p
    XMLor xml_tuple5 ''OC3'' oc oc oc s s  OC3
    XMLor xml_tuple5 ''OC3prime'' oc oc oc s s OC3p
    XMLor xml_pair ''OCintoDP1'' dp oc OCDP1
    XMLor xml_pair ''OCintoDP2'' dp oc OCDP2
    XMLor xml_pair ''equivalent'' dp dp WPEQ
    XMLor xml_pair ''lift'' dp dp Lift
    XMLor xml_tuple5 ''DP_OC_1_1'' dp dp oc s s DPOC1_1
    XMLor xml_tuple6 ''DP_OC_1_2'' dp dp oc s s s DPOC1_2
    XMLor xml_tuple5 ''DP_OC_2'' dp dp oc s s DPOC2
    XMLor xml_tuple5 ''DP_OC_3_1'' dp dp oc s s DPOC3_1
    XMLor xml_tuple6 ''DP_OC_3_2'' dp dp oc s s s DPOC3_2
    XMLor xml_tuple5 ''DP_DP_1_1'' dp dp dp s s DPDP1_1
    XMLor xml_tuple5 ''DP_DP_1_2'' dp dp dp s s DPDP1_2
    XMLor xml_tuple5 ''DP_DP_2_1'' dp dp dp s s DPDP2_1
    XMLor xml_tuple5 ''DP_DP_2_2'' dp dp dp s s DPDP2_2
  ) id)"

definition nonloop_srs_reason :: "(_ dp_proof_step list \<Rightarrow> _ non_loop_srs_proof)xmlt2" where
  "nonloop_srs_reason \<equiv>
    xml_triple ''selfEmbeddingOC''
        string string string
        (\<lambda> l m r. SE_OC (m, l @ m @ r) l r)
    XMLor xml_triple ''selfEmbeddingDP''
        derivation_pattern string string
        SE_DP"

definition nonloop_srs :: "_ non_loop_srs_proof xmlt2" where
  "nonloop_srs \<equiv> xml_pair ''nonterminatingSRS''
    (xml_many ''derivationPatterns'' derivation_pattern_proof id)
    nonloop_srs_reason
    (\<lambda> list prf. prf list)"

definition not_wn_ta :: "(_, string) not_wn_ta_prf xmlt2" where
  "not_wn_ta \<equiv> xml_pair ''notWNTreeAutomaton''
    (tree_automaton (ta_normal_lhs xml2name))
    closed_criterion
    (\<lambda> ta rel. Not_wn_ta_prf ta rel)"

definition xml2inn_fp_nt_trs_assm ::
    "(('f, 'l, string) qtrsLL + ('f, 'l, string) fptrsLL) xmlt2"
where
  "xml2inn_fp_nt_trs_assm \<equiv> xml_change (xml2_trs_input xml2name) (\<lambda> inp.
   case inp of
     Inn_TRS_input inn r s start \<Rightarrow>
      if s \<noteq> [] then
        xml_error ''relative rules are not allowed here''
      else if start \<noteq> Full then
        xml_error ''start terms not allowed here''
      else
        xml_return (Inl (default_nfs_nt_trs, strategy_to_Q inn r, r))
    | FP_TRS_input fp r \<Rightarrow> xml_return (Inr (strategy_to_fp fp r, r))
    | _ \<Rightarrow> xml_error ''non-relative TRS expected''
  )"

definition xml2fp_nt_trs_assm :: "('f, 'l, string) fptrsLL xmlt2"
where
  "xml2fp_nt_trs_assm \<equiv> xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of
      FP_TRS_input fp r \<Rightarrow> xml_return (strategy_to_fp fp r, r)
    | _ \<Rightarrow> xml_error ''outermost/forbidden-pattern/context-sensitive TRS expected''
  )"

definition xml2inn_nt_trs_assm where
  "xml2inn_nt_trs_assm \<equiv> xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of Inn_TRS_input inn r s start \<Rightarrow>
      if s \<noteq> [] then xml_error ''(innermost) TRS without relative rules expected''
      else if start \<noteq> Full then xml_error ''start term is not allowed here''
      else xml_return (default_nfs_nt_trs, strategy_to_Q inn r, r)
   )"

definition xml2inn_rel_nt_trs_assm where
  "xml2inn_rel_nt_trs_assm \<equiv> xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of Inn_TRS_input inn r s start \<Rightarrow>
    if start \<noteq> Full then xml_error ''start term is not allowed here''
    else xml_return (default_nfs_nt_trs, strategy_to_Q inn r, r, s)
   )"
(*
abbreviation sub_disproof where "sub_disproof xml2name
  xml2trs_nontermination_proof
  xml2dp_nontermination_proof
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof
  x \<equiv>
    let cs = Xml.children x
     in (if length cs \<noteq> 2 then Xmlt.fail ''subProof'' x
      else let inp = take 2 (Xml.tag (cs ! 1)) in
      (if inp = ''tr'' then
        do {
          io_trs <- xml2inn_fp_nt_trs_assm xml2name (hd cs);
          (case io_trs of
            Inl qtrs \<Rightarrow> Xmlt.change xml2trs_nontermination_proof (\<lambda> prf. Not_SN_assm_proof qtrs prf :: ('f, 'l, string, 'a, 'b, 'c, 'd,'e) generic_assm_proof)
          | Inr fptrs \<Rightarrow> Xmlt.change xml2fp_nontermination_proof (\<lambda> prf. Not_SN_FP_assm_proof fptrs prf)) (cs ! 1)
        }
      else if inp = ''dp'' then
        Xmlt.pair ''subProof'' (xml2dp_input' False xml2name) xml2dp_nontermination_proof (\<lambda> qdp prf. Infinite_assm_proof qdp prf :: ('f, 'l, string, 'a, 'b, 'c, 'd,'e) generic_assm_proof) x
      else if inp = ''re'' then
        Xmlt.pair ''subProof'' (xml2inn_rel_nt_trs_assm xml2name) xml2reltrs_nontermination_proof Not_RelSN_assm_proof x
      else if inp = ''un'' then
        Xmlt.pair ''subProof'' xml2unknown_input xml2unknown_disproof Unknown_assm_proof x
      else Xmlt.fail ''subProof'' x))"
*)

function xml2dp_nontermination_proof ::
    "('f, 'l, string) dp_nontermination_proof xmlt2" and
  xml2trs_nontermination_proof ::
    "('f, 'l, string) trs_nontermination_proof xmlt2" and
  xml2reltrs_nontermination_proof ::
    "('f, 'l, string) reltrs_nontermination_proof xmlt2" and
  xml2unknown_disproof :: "('f, 'l, string) neg_unknown_proof xmlt2"
  where
  "xml2dp_nontermination_proof x = xml_singleton ''dpNonterminationProof'' (
    XMLdo ''dpRuleRemoval'' {
      p \<leftarrow>? xml_singleton ''dps'' (XTC_Parser.rules xml2name) id;
      r \<leftarrow>? xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Rule_Removal (Rule_removal_nonterm_dp_prf p r) prf)
    } XMLor xml_change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). xml_return (DP_Loop (DP_loop_prf s rseq \<sigma> C)))
    XMLor xml_change (CPF_Parser.nonloop xml2name)
      (xml_return \<circ> DP_Nonloop)
    XMLor xml_pair ''innermostLhssRemovalProc''
      (innermostLhss xml2name)
      xml2dp_nontermination_proof
      (\<lambda> q. DP_Q_Reduction (DP_q_reduction_nonterm_prf q))
    XMLor xml_pair ''innermostLhssIncreaseProc''
      (innermostLhss xml2name)
      xml2dp_nontermination_proof
      (\<lambda> q. DP_Q_Increase (Q_increase_nonterm_dp_prf q))
    XMLor xml_pair ''instantiationProc''
      (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
      xml2dp_nontermination_proof
      (\<lambda> p. DP_Instantiation (Instantiation_complete_proc_prf p))
    XMLor xml_tuple4 ''narrowingProc''
      (XTC_Parser.rule xml2name)
      XTC_Parser.pos
      (xml_singleton ''narrowings'' (XTC_Parser.rules xml2name) id)
      xml2dp_nontermination_proof
      (\<lambda> st po p. DP_Narrowing (Narrowing_complete_proc_prf st po p))
    XMLor XMLdo ''rewritingProc'' {
             (s,t) \<leftarrow> XTC_Parser.rule xml2name;
             (p,lr,t') \<leftarrow> CPF_Parser.rstep xml2name;
             c \<leftarrow>? XTC_Parser.rule xml2name;
             Uopt \<leftarrow>? xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id;
             prf \<leftarrow> xml2dp_nontermination_proof;
             let st'' = (case c of Some v \<Rightarrow> v | None \<Rightarrow> (s,t'));
             xml_return (DP_Rewriting (Rewriting_complete_proc_prf Uopt (s,t) (s,t') st'' lr p) prf)
    } XMLor xml_pair ''switchFullStrategyProc''
      (CPF_Parser.wcr_proof xml2name)
      xml2dp_nontermination_proof
      DP_Termination_Switch
    XMLor xml_singleton ''infinitenessAssumption''
      (xml2dp_input' xml2name False)
      (\<lambda> qdp. DP_Assume_Infinite qdp [])
    XMLor XMLdo ''unknownProof'' {
      a \<leftarrow> xml_text ''description'';
      b \<leftarrow> xml2dp_input' xml2name False;
      c \<leftarrow>* xml_any;
      xml_return (DP_Assume_Infinite b [])
  }) id x"
| "xml2trs_nontermination_proof x = xml_singleton ''trsNonterminationProof'' (
  xml_leaf ''variableConditionViolated'' TRS_Not_Well_Formed
  XMLor xml_change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). xml_return (TRS_Loop (TRS_loop_prf s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)))
  XMLor xml_change (CPF_Parser.nonloop xml2name) (xml_return \<circ> TRS_Nonloop)
  XMLor xml_change (CPF_Parser.nonloop_srs xml2name) (xml_return \<circ> TRS_Nonloop_SRS)
  XMLor xml_pair ''ruleRemoval''
    (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
    xml2trs_nontermination_proof
    (\<lambda> r. TRS_Rule_Removal (Rule_removal_nonterm_trs_prf r))
  XMLor xml_triple ''dpTrans''
    (xml_singleton ''dps'' (XTC_Parser.rules xml2name) id)
    (xml_bool ''markedSymbols'')
    xml2dp_nontermination_proof
    (\<lambda>p _. TRS_DP_Trans (DP_trans_nontermination_tt_prf p))
  XMLor xml_pair ''stringReversal''
    (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
    xml2trs_nontermination_proof
    (\<lambda>_. TRS_String_Reversal)
  XMLor xml_tuple4 ''constantToUnary''
     (XTC_Parser.plain_var)
     (CPF_Parser.renaming xml2name)
     (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
     xml2trs_nontermination_proof
     (\<lambda> v ren S p. TRS_Constant_String (Const_string_complete_proof v ren S) p)
  XMLor xml_pair ''innermostLhssIncrease''
    (innermostLhss xml2name)
    xml2trs_nontermination_proof
    (\<lambda> q. TRS_Q_Increase (Q_increase_nonterm_trs_prf q))
  XMLor xml_pair ''switchFullStrategy''
    (CPF_Parser.wcr_proof xml2name)
    xml2trs_nontermination_proof
    TRS_Termination_Switch
  XMLor xml_triple ''uncurry''
      (CPF_Parser.uncurry_info xml2name)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
      xml2trs_nontermination_proof
      (\<lambda> i r. TRS_Uncurry (Uncurry_nt_proof i r))
  XMLor xml_change (CPF_Parser.not_wn_ta xml2name)
    (xml_return \<circ> TRS_Not_WN_Tree_Automaton)
  XMLor xml_singleton ''nonterminationAssumption''
    xml2inn_nt_trs_assm
    (\<lambda> qtrs. TRS_Assume_Not_SN qtrs [])
  XMLor XMLdo ''unknownProof'' {
    a \<leftarrow> xml_text ''description'';
    b \<leftarrow> xml2inn_nt_trs_assm;
    c \<leftarrow>* xml_any;
    xml_return (TRS_Assume_Not_SN b [])
  }) id x"
| "xml2reltrs_nontermination_proof x = xml_singleton ''relativeNonterminationProof'' (
  xml_leaf ''variableConditionViolated'' Rel_Not_Well_Formed
  XMLor xml_change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). xml_return (Rel_Loop (Rel_trs_loop_prf s rseq \<sigma> C)))
  XMLor xml_change xml2trs_nontermination_proof (xml_return \<circ> Rel_R_Not_SN)
  XMLor xml_triple ''ruleRemoval''
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) Some)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) Some)
      xml2reltrs_nontermination_proof
      (\<lambda> r s. Rel_Rule_Removal (Rule_removal_nonterm_reltrs_prf r s))
  XMLor xml_triple ''stringReversal''
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) Some)
      (xml_singleton ''trs'' (XTC_Parser.rules xml2name) Some)
      xml2reltrs_nontermination_proof
      (\<lambda> _ _. Rel_TRS_String_Reversal)
  XMLor xml_singleton ''nonterminationAssumption''
      xml2inn_rel_nt_trs_assm
      (\<lambda> qtrs. Rel_TRS_Assume_Not_SN qtrs [])
  XMLor XMLdo ''unknownProof'' {
      a \<leftarrow> xml_text ''description'';
      b \<leftarrow> xml2inn_rel_nt_trs_assm;
      c \<leftarrow>* xml_any;
      xml_return (Rel_TRS_Assume_Not_SN b [])
  }) id x"
| "xml2unknown_disproof x = xml_singleton ''unknownInputProof'' (
    xml_singleton ''unknownAssumption''
        xml2unknown_input
        (\<lambda> u. Assume_NT_Unknown u [])
    XMLor XMLdo ''unknownProof'' {
        a \<leftarrow> xml_text ''description'';
        b \<leftarrow> xml2unknown_input;
        c \<leftarrow>* xml_any;
        xml_return (Assume_NT_Unknown b [])
  }) id x"
by (pat_completeness, auto)

termination by (relation "measure (\<lambda>x. let y = case x of Inl(Inl y) \<Rightarrow> y | Inl (Inr y) \<Rightarrow> y | Inr (Inl y) \<Rightarrow> y | Inr (Inr y) \<Rightarrow> y in size_xml_state y)", auto)


definition rule_labeling_function :: "(_, string) rule_lab_repr xmlt2"
where
  "rule_labeling_function \<equiv> xml_many ''ruleLabelingFunction''
   (xml_pair ''ruleLabelingFunctionEntry''
    (XTC_Parser.rule xml2name)
    (xml_nat ''label'')
    Pair) id"

fun add_source_lab_proof
where
  "add_source_lab_proof (Rule_Labeling rl js _) _ prf = Rule_Labeling rl js (Some prf)"
| "add_source_lab_proof (Rule_Labeling_Conv rl cs _) (Some n) prf = Rule_Labeling_Conv rl cs (Some (n, prf))"

fun nat_or_empty (*TODO: ad-hoc, syntax should be changed*)
where "nat_or_empty ([],s) = Inr None"
| "nat_or_empty ([XML_text txt],s) = (
  case int_of_string txt of Inr n \<Rightarrow>
    Inr (Some (Int.nat n))
  | Inl err \<Rightarrow> xml_error err ([]::xml list,s)
  )"

definition xml2ms_signature
  where "xml2ms_signature = xml_many ''manySortedSignature'' (
       xml_triple ''manySortedFunction''
         xml2name
         (xml_many ''args'' (xml_text ''sort'') id)
         (xml_singleton ''result'' (xml_text ''sort'') id)
       (\<lambda>a b c. (a, b, c))
    ) id"

function xml2cr_proof :: "(_,_,string) cr_proof xmlt2"
where "xml2cr_proof x = xml_singleton ''crProof'' (
    xml_pair ''wcrAndSN''
        (wcr_proof xml2name)
        xml2trs_termination_proof
      SN_WCR
    XMLor xml_leaf ''orthogonal'' Weakly_Orthogonal
    XMLor xml_change (xml_nat ''stronglyClosed'') (xml_return \<circ> Strongly_Closed)
    XMLor xml_change (xml_do ''parallelClosed'' nat_or_empty) (xml_return \<circ> Parallel_Closed)
    XMLor xml_triple ''criticalPairClosingSystem''
        (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
        xml2trs_termination_proof
        (xml_nat ''nrSteps'')
      Critical_Pair_Closing_System
    XMLor xml_pair ''ruleLabeling''
        (CPF_Parser.rule_labeling_function xml2name)
        (CPF_Parser.joinable_critical_pairs xml2name)
      (\<lambda> rl js. Rule_Labeling rl js None)
    XMLor XMLdo ''decreasingDiagrams'' {
        a \<leftarrow>? xml2trs_termination_proof;
        ret \<leftarrow>
          XMLdo ''ruleLabeling'' {
            rl \<leftarrow> CPF_Parser.rule_labeling_function xml2name;
            js \<leftarrow> CPF_Parser.joinable_critical_pairs xml2name;
            case a of
              Some prf \<Rightarrow> xml_return (add_source_lab_proof (Rule_Labeling rl js None) None prf)
            | _ \<Rightarrow> xml_return (Rule_Labeling rl js None)
          } XMLor XMLdo ''ruleLabelingConv'' {
            rl \<leftarrow> CPF_Parser.rule_labeling_function xml2name;
            cs \<leftarrow> CPF_Parser.convertible_critical_peaks xml2name;
            case a of
              Some prf \<Rightarrow> XMLdo {
                n \<leftarrow> xml_nat ''nrSteps'';
                xml_return (add_source_lab_proof (Rule_Labeling_Conv rl cs None) (Some n) prf)
              }
            | _ \<Rightarrow> xml_return (Rule_Labeling_Conv rl cs None)
          };
        xml_return ret
    } XMLor XMLdo ''redundantRules'' {
      trs \<leftarrow> xml_singleton ''trs'' (XTC_Parser.rules xml2name) id;
      n \<leftarrow> xml_nat ''nrSteps'';
      cs \<leftarrow>[[]] xml_many ''conversions'' (CPF_Parser.conversion xml2name) (map snd);
      prf \<leftarrow> xml2cr_proof;
      xml_return (Redundant_Rules trs n cs prf)
    } XMLor XMLdo ''persistentDecomposition'' {
      sig \<leftarrow> xml2ms_signature;
      subs \<leftarrow>* xml_pair ''component''
       (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
       xml2cr_proof
       Pair;
      xml_return (Persistent_Decomposition sig subs)
  }) id x"
by (pat_completeness, auto)

termination by (relation "measure size_xml_state", auto)

definition compatible_ta :: "string \<Rightarrow> ((string,_)tree_automaton \<times> string ta_relation) xmlt2" where
  "compatible_ta tag \<equiv> XMLdo tag {
    a \<leftarrow> tree_automaton (ta_normal_lhs xml2name);
    b \<leftarrow>[Id_Relation] closed_criterion;
    xml_return (a,b)
  }"

definition xml2completion_proof :: "(_,_,string) completion_proof xmlt2"
where "xml2completion_proof \<equiv> XMLdo ''completionProof'' {
     w \<leftarrow> wcr_proof xml2name;
     t \<leftarrow> xml2trs_termination_proof;
     (s1,s2) \<leftarrow> XMLdo ''equivalenceProof'' {
       s1 \<leftarrow> CPF_Parser.subsumption_proof xml2name;
       s2 \<leftarrow>? CPF_Parser.subsumption_proof xml2name;
       xml_return (s1,s2)
      };
     xml_return (SN_WCR_Eq w t s1 s2)
  }"

definition xml2ordered_completion_step :: "(_,string) oc_irule xmlt2"
  where "xml2ordered_completion_step \<equiv>
      xml_triple ''deduce''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda> s t u. OC_Deduce s t u)
    XMLor 
      (xml_pair ''orientl''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t. OC_Orientl s t))
    XMLor 
      (xml_pair ''orientr''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t. OC_Orientr s t))
    XMLor 
      (xml_singleton ''delete''
        (XTC_Parser.term xml2name)
        (\<lambda>s. OC_Delete s))
    XMLor 
      (xml_triple ''compose''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Compose s t u))
    XMLor 
      (xml_triple ''simplifyl''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Simplifyl s t u))
    XMLor 
      (xml_triple ''simplifyr''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Simplifyr s t u))
    XMLor 
      (xml_triple ''collapse''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Collapse s t u))
"

definition xml2ordered_completion_proof :: "(_,_,string) ordered_completion_proof xmlt2"
  where "xml2ordered_completion_proof \<equiv> 
           xml_singleton ''orderedCompletionProof'' (
             xml_many ''run''
               (xml_singleton ''orderedCompletionStep'' xml2ordered_completion_step id)
               id
           ) OKB"

definition "unraveling_info \<equiv>
  (xml_many ''unravelingInformation'' (XMLdo ''unravelingEntry'' {
     a \<leftarrow> XTC_Parser.crule xml2name ''conditionalRule'';
     b \<leftarrow>* XTC_Parser.rule xml2name;
     xml_return (a,b)
  }) id)"



definition xml2quasi_reductive_proof :: "(_,_,string) quasi_reductive_proof xmlt2"
where
  "xml2quasi_reductive_proof \<equiv> xml_singleton ''quasiReductiveProof'' (
    xml_pair ''unraveling''
        (CPF_Parser.unraveling_info xml2name)
        xml2trs_termination_proof
        Unravel
  ) id"

definition "xml2state_map \<equiv> xml_many ''stateMap'' (xml_pair ''entry''
  state
  (XTC_Parser.term xml2name)
  Pair)
  (\<lambda>xs. \<lambda>x. the (map_of xs x))"

definition xml2nonreachable_etac_info
where
  "xml2nonreachable_etac_info \<equiv> xml_tuple5 ''nonreachableEtac''
    (CPFsignature xml2name)
    xml2name
    xml2name
    (tree_automaton (ta_normal_lhs xml2name))
    (xml2state_map)
    (\<lambda>F a c A m. Nonreachable_ETAC F a c (map_states_impl m A))"

function xml2nonreachability_proof :: "(_, _) nonreachability_proof xmlt2"
where "xml2nonreachability_proof x = xml_singleton ''nonreachabilityProof'' (
    xml_leaf ''nonreachableTcap'' Nonreachable_Gtcap
    XMLor xml2nonreachable_etac_info
    XMLor xml_pair ''nonreachableSubstApprox''
      (XTC_Parser.rules xml2name)
      xml2nonreachability_proof
      Nonreachable_Subst_Approx
    XMLor xml_singleton ''nonreachableReverse'' xml2nonreachability_proof
      Nonreachable_Reverse
  ) id x"
by (pat_completeness,auto)

termination by (relation "measure size_xml_state", auto)


definition "xml2nonjoinability_proof \<equiv> xml_singleton ''nonjoinabilityProof'' (
  xml_leaf ''nonjoinableTcap'' Nonjoinable_Tcap
  XMLor xml_change xml2nonreachability_proof (xml_return \<circ> Nonjoinable_Ground_NF)
  ) id"

function xml2infeasibility_proof :: "(_,_) infeasibility_proof xmlt2"
where "xml2infeasibility_proof x = xml_singleton ''infeasibilityProof'' (
    xml_pair ''infeasibleCompoundConditions''
      xml2name
      xml2nonreachability_proof
      Infeasible_Compound_Conditions
    XMLor xml_pair ''infeasibleEquation''
      (XTC_Parser.rule xml2name)
      xml2nonreachability_proof
      (\<lambda>(l, r). Infeasible_Equation l r)
    XMLor xml_pair ''infeasibleSubset''
      (XTC_Parser.rules xml2name)
      xml2infeasibility_proof
      Infeasible_Subset
    XMLor xml_tuple4 ''infeasibleRhssEqual''
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      xml2nonjoinability_proof
      Infeasible_Rhss_Equal
  ) id x"
by (pat_completeness,auto)

termination by (relation "measure size_xml_state", auto)


definition "xml2ao_infeasibility_proof \<equiv> xml_singleton ''aoInfeasibilityProof'' (
  xml_change xml2infeasibility_proof (xml_return \<circ> AO_Infeasibility_Proof)
  XMLor xml_tuple4 ''aoLhssEqual''
    (XTC_Parser.term xml2name)
    (XTC_Parser.term xml2name)
    (XTC_Parser.term xml2name)
    xml2nonjoinability_proof
    AO_Lhss_Equal
  ) id"

definition xml2infeasible_conds
where
  "xml2infeasible_conds \<equiv> xml_many ''infeasibleConditions''
    (xml_pair ''infeasibleCondition''
      (XTC_Parser.conditions xml2name ''conditions'')
      xml2infeasibility_proof
      Pair)
    id"

definition xml2ao_infeasible_conds
where
  "xml2ao_infeasible_conds \<equiv> xml_many ''aoInfeasibleConditions''
    (xml_triple ''aoInfeasibleCondition''
      (XTC_Parser.conditions xml2name ''conditions'')
      (XTC_Parser.conditions xml2name ''conditions'')
      xml2ao_infeasibility_proof
      (\<lambda>cs\<^sub>1 cs\<^sub>2 p. (cs\<^sub>1, cs\<^sub>2, p)))
    id"

function cstep :: "(_, _) cstep_proof xmlt2"
     and csteps :: "(_, _) cstep_proof list xmlt2"
where
  "cstep x = xml_tuple6 ''conditionalRewriteStep''
    (XTC_Parser.crule xml2name ''rule'')
    XTC_Parser.pos
    (CPF_Parser.subst xml2name)
    (xml_singleton ''source'' (XTC_Parser.term xml2name) id)
    (xml_singleton ''target'' (XTC_Parser.term xml2name) id)
    (xml_many ''conditions'' csteps id)
    (\<lambda>r p s t u qs. Cstep_step r p (mk_subst Var s) t u qs) x"
| "csteps x = xml_many ''conditionalRewritingSequence'' cstep id x"
by (pat_completeness,auto)

termination by (relation "measure (\<lambda>x. case x of Inl x \<Rightarrow> size_xml_state x | Inr x \<Rightarrow> size_xml_state x)", auto)

definition xml2const_map :: "(_ \<Rightarrow> string option) xmlt2"
where
  "xml2const_map \<equiv> xml_many ''constMap'' (xml_pair ''entry''
    (xml_singleton ''symbol'' xml2name id)
    (xml_text ''const'')
    Pair)
    (\<lambda>xs. map_of xs)"

definition xml2context_joinable_proof :: "(_, _) context_joinable_proof xmlt2"
where
  "xml2context_joinable_proof \<equiv> xml_tuple4 ''contextJoinabilityProof''
    (XTC_Parser.term xml2name)
    (CPF_Parser.csteps xml2name)
    (CPF_Parser.csteps xml2name)
    xml2const_map
    (\<lambda>t ps qs m. Contextual_Join (orig_term m t) (map (orig_cstep m) ps) (map (orig_cstep m) qs))"

definition xml2context_joinable_ccps ::
  "((_, string) term \<times> (_, string) term \<times> (_, string) rules \<times> (_, string) context_joinable_proof) list xmlt2"
where
  "xml2context_joinable_ccps \<equiv> XMLdo ''contextJoinableCCPs'' {
    ret \<leftarrow>* XMLdo ''contextJoinableCCP'' {
      s \<leftarrow> XTC_Parser.term xml2name;
      t \<leftarrow> XTC_Parser.term xml2name;
      cs \<leftarrow> XTC_Parser.conditions xml2name ''conditions'';
      p \<leftarrow> xml2context_joinable_proof;
      xml_return (s, t, cs, p)
    };
    xml_return ret
  }"

definition xml2unfeasible_proof :: "(_, string) unfeasible_proof xmlt2"
where
  "xml2unfeasible_proof \<equiv> xml_tuple5 ''unfeasibilityProof''
    (xml_triple ''terms''
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
    (\<lambda>t u v. (t, u, v)))
    csteps
    csteps
    (xml_pair ''rules'' (XTC_Parser.crule xml2name ''rule'') (XTC_Parser.crule xml2name ''rule'') Pair)
    xml2const_map
    (\<lambda>(t, u, v) ps qs (r, r') m.
      UnfeasibleOverlap (orig_term m t) (orig_term m u) (orig_term m v)
        (map (orig_cstep m) ps) (map (orig_cstep m) qs) r r')"

definition xml2unfeasible_ccps ::
  "((_, string) subst \<times> (_, string) unfeasible_proof) list xmlt2"
where
  "xml2unfeasible_ccps \<equiv> xml_many ''unfeasibleCCPs''
    (xml_pair ''unfeasibleCCP''
      (CPF_Parser.subst xml2name)
      xml2unfeasible_proof
      (\<lambda>s p. (subst_of s, p)))
    id"

definition xml2inline_cond_info
  where "xml2inline_cond_info \<equiv> XMLdo ''inlinedRules'' {
    ret \<leftarrow>* XMLdo ''inlinedRule'' {
      rule \<leftarrow> XTC_Parser.crule xml2name ''rule'';
      conds \<leftarrow> XTC_Parser.conditions xml2name ''inlinedConditions'';
      xml_return (rule,conds)
    };
    xml_return ret
  }"

definition xml2infeasible_rules_info
  where
    "xml2infeasible_rules_info \<equiv> xml_many ''infeasibleRules''
      (xml_pair ''infeasibleRule''
        (XTC_Parser.crule xml2name ''rule'')
        xml2infeasibility_proof
        Pair)
      id"

function xml2conditional_cr_proof ::
    "(_, _, string) conditional_cr_proof xmlt2"
where "xml2conditional_cr_proof x = xml_singleton ''conditionalCrProof'' (
    xml_singleton ''unconditional''
      xml2cr_proof
      Unconditional_CR
    XMLor xml_pair ''unraveling''
      (CPF_Parser.unraveling_info xml2name)
      xml2cr_proof
      Unravel_CR
    XMLor xml_triple ''inlineConditions''
      (XMLdo ''rules'' { ret \<leftarrow>* XTC_Parser.crule xml2name ''rule''; xml_return ret })
      xml2inline_cond_info
      xml2conditional_cr_proof
      (Transformation_CR \<circ>\<circ> Inline_Conditions_CCRT)
    XMLor xml_pair ''infeasibleRuleRemoval''
      xml2infeasible_rules_info
      xml2conditional_cr_proof
      (Transformation_CR \<circ> Infeasible_Rule_Removal_CCRT)
    XMLor xml_leaf ''almostOrthogonal'' Almost_Orthogonal_CR
    XMLor xml_singleton ''almostOrthogonalModuloInfeasibility''
        xml2ao_infeasible_conds Almost_Orthogonal_Modulo_Infeasibility_CR'
    XMLor xml_tuple4 ''al94''
      xml2quasi_reductive_proof
      xml2context_joinable_ccps
      xml2infeasible_conds
      xml2unfeasible_ccps
      AL94_CR
  ) id x"
by (pat_completeness,auto)

termination by (relation "measure size_xml_state", auto)

end

hide_const (open) string
hide_const (open) nonloop_srs
hide_const (open) rule_labeling_function

function xml2non_join_info :: "string lt xmlt2 \<Rightarrow> (string lt,string,string) non_join_info xmlt2" where
  "xml2non_join_info xml2name x = (
    xml_leaf ''distinctNormalForms'' Diff_NFs
    XMLor xml_leaf ''capNotUnif'' Tcap_Not_Unif
    XMLor xml_pair ''subterm''
        XTC_Parser.pos
        (xml2non_join_info xml2name)
        Subterm_NJ
    XMLor xml_pair ''grounding''
        (CPF_Parser.subst xml2name)
        (xml2non_join_info xml2name)
        Grounding
    XMLor xml_pair ''emptyTreeAutomataIntersection''
        (compatible_ta xml2name ''firstAutomaton'')
        (compatible_ta xml2name ''secondAutomaton'')
        (\<lambda> (ta1,rel1) (ta2,rel2). Tree_Aut_Intersect_Empty ta1 rel1 ta2 rel2)
    XMLor xml_singleton ''differentInterpretation''
        (xmlt2 (CPF_Parser.sl_variant (xmlt_of_xmlt2 xml2name)))
        Finite_Model_Gt
    XMLor xml_singleton ''strictDecrease''
        (CPF_Parser.ordering_constraint_proof xml2name False)
        Reduction_Pair_Gt
    XMLor xml_pair ''argumentFilterNonJoin''
        (CPF_Parser.afs xml2name)
        (xml2non_join_info xml2name)
        Argument_Filter_NJ
    XMLor XMLdo ''usableRulesNonJoin'' {
      a \<leftarrow>? xml_leaf ''left'' True XMLor xml_leaf ''right'' False;
      case a of None \<Rightarrow> XMLdo {
          b \<leftarrow> xml2non_join_info xml2name;
          xml_return (Usable_Rules_Reach_NJ b)
        }
      | Some left \<Rightarrow> XMLdo {
          U \<leftarrow> xml_singleton ''usableRules'' (XTC_Parser.rules xml2name) id;
          p \<leftarrow> xml2non_join_info xml2name;
          xml_return (Usable_Rules_Reach_Unif_NJ (if left then Inl U else Inr U) p)
        }
    }
  ) x"
by (pat_completeness,auto)

termination by (relation "measure (size_xml_state \<circ> snd)", auto)

function xml2ncr_proof :: "(string,label_type)lab xmlt2 \<Rightarrow> (string, label_type,string,string) ncr_proof xmlt2"
where "xml2ncr_proof xml2name x = (let rew = CPF_Parser.rsteps xml2name in
    xml_singleton ''crDisproof'' (
      XMLdo ''nonWcrAndSN'' {
       _ \<leftarrow> xml_any;
       prf \<leftarrow> xml2trs_termination_proof xml2name;
       xml_return (SN_NWCR prf)
    } XMLor xml_triple ''nonJoinableFork'' rew rew
           (xml2non_join_info xml2name)
           (\<lambda> (s,seq1) (_,seq2). Non_Join s seq1 seq2)
    XMLor xml_pair ''modularityDisjoint''
        (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
        (xml2ncr_proof xml2name)
        NCR_Disj_Subtrs
    XMLor xml_triple ''redundantRules''
        (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
        (xml_nat ''nrSteps'')
        (xml2ncr_proof xml2name)
      NCR_Redundant_Rules
    XMLor xml_pair ''persistentDecomposition''
        (xml2ms_signature xml2name)
        (xml_pair ''component''
            (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
            (xml2ncr_proof xml2name)
            (\<lambda> trs prf. (trs, prf)))
       (\<lambda> sig (trs, prf). NCR_Persistent_Decomposition sig trs prf)
  ) id x)"
by (pat_completeness, auto)

termination by (relation "measure (size_xml_state \<circ> snd)", auto)

hide_const(open) unraveling_info

function xml2conditional_ncr_proof ::
    "(string, label_type) lab xmlt2 \<Rightarrow> (string, label_type, string, string) conditional_ncr_proof xmlt2"
where "xml2conditional_ncr_proof xml2name x = xml_singleton ''conditionalCrDisproof'' (
    xml_singleton ''unconditional''
      (CPF_Parser.xml2ncr_proof xml2name)
      Unconditional_CNCR
    XMLor xml_triple ''inlineConditions''
      (XMLdo ''rules'' {ret \<leftarrow>* XTC_Parser.crule xml2name ''rule''; xml_return ret})
      (xml2inline_cond_info xml2name)
      (xml2conditional_ncr_proof xml2name)
      (Transformation_CNCR \<circ>\<circ> Inline_Conditions_CCRT)
    XMLor xml_pair ''infeasibleRuleRemoval''
      (xml2infeasible_rules_info xml2name)
      (xml2conditional_ncr_proof xml2name)
      (Transformation_CNCR \<circ> Infeasible_Rule_Removal_CCRT)
    XMLor xml_tuple4 ''nonJoinableFork''
      (xml_triple ''terms''
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
      (\<lambda>s t u. (s, t, u)))
      (CPF_Parser.csteps xml2name)
      (CPF_Parser.csteps xml2name)
      (xml2non_join_info xml2name)
      (\<lambda>(s, t, u) ps qs i. Non_Join_CNCR s t u ps qs i)
  ) id x"
by (pat_completeness,auto)

termination by (relation "measure (size_xml_state \<circ> snd)", auto)

function xml2eq_proof :: "'a::show xmlt2 \<Rightarrow> ('a, string) eq_proof xmlt2" where
  "xml2eq_proof xml2name x = (
    xml_singleton ''refl'' (XTC_Parser.term xml2name) Equational_Reasoning.Refl
    XMLor xml_singleton ''sym'' (xml2eq_proof xml2name) Sym
    XMLor xml_pair ''trans'' (xml2eq_proof xml2name) (xml2eq_proof xml2name) Trans
    XMLor xml_pair ''assm'' (XTC_Parser.rule xml2name) (CPF_Parser.subst xml2name) (\<lambda>r s. Assm r (mk_subst Var s))
    XMLor XMLdo ''cong'' {
      a \<leftarrow> xml2name;
      b \<leftarrow>* xml2eq_proof xml2name;
      xml_return (Cong a b)
   }) x"
by (pat_completeness,auto)

termination by (relation "measure (size_xml_state \<circ> snd)", auto)

definition
  xml2equational_proof :: "('f :: {compare_order,show}, label_type) lab xmlt2 \<Rightarrow> ('f, label_type, string) equational_proof xmlt2"
where
  "xml2equational_proof xml2name =
    xml_singleton ''equationalProof'' (
      xml_singleton ''equationalProofTree'' (xml2eq_proof xml2name) Equational_Proof_Tree
      XMLor xml_change (CPF_Parser.conversion xml2name) (\<lambda>(_, e). xml_return (Conversion e))
      XMLor xml_change (CPF_Parser.subsumption_proof xml2name) (xml_return \<circ> Conversion_With_History)
      XMLor xml_pair ''completionAndNormalization''
        (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization
    ) id"

definition
  xml2equational_disproof :: "('f :: {compare_order,show}, label_type) lab xmlt2 \<Rightarrow> ('f, label_type, string) equational_disproof xmlt2"
where
  "xml2equational_disproof xml2name =
    xml_singleton ''equationalDisproof'' (
      xml_pair ''completionAndNormalization''
        (xml_singleton ''trs'' (XTC_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization_Different
      XMLor xml_singleton ''convertibleInstance'' (CPF_Parser.subsumption_proof xml2name) Convertible_Instance
    ) id"

type_synonym lsymbol = "(string, label_type) lab"

definition xml2cert_problem :: "string xmlt2 \<Rightarrow> lsymbol xmlt2 \<Rightarrow> _ xmlt2"
where
  "xml2cert_problem xml2uname xml2name \<equiv> XMLdo ''certificationProblem'' {
    inp \<leftarrow>? CPF_Parser.xml2input xml2name;
    version \<leftarrow>? xml_text ''cpfVersion'';
    p \<leftarrow> XMLdo ''proof'' {
      p \<leftarrow> xml_change (xml2trs_termination_proof xml2name) (xml_return \<circ> TRS_Termination_Proof)
        XMLor xml_change (xml2trs_nontermination_proof xml2name) (xml_return \<circ> TRS_Nontermination_Proof)
        XMLor xml_change (xml2reltrs_nontermination_proof xml2name) (xml_return \<circ> Relative_TRS_Nontermination_Proof)
        XMLor xml_change (xml2cr_proof xml2name) (xml_return \<circ> TRS_Confluence_Proof)
        XMLor xml_change (xml2ncr_proof xml2name) (xml_return \<circ> TRS_Non_Confluence_Proof)
        XMLor xml_change (xml2dp_termination_proof xml2name) (xml_return \<circ> DP_Termination_Proof)
        XMLor xml_change (xml2dp_nontermination_proof xml2name) (xml_return \<circ> DP_Nontermination_Proof)
        XMLor xml_change (xml2completion_proof xml2name) (xml_return \<circ> Completion_Proof)
        XMLor xml_change (xml2ordered_completion_proof xml2name) (xml_return \<circ> Ordered_Completion_Proof)
        XMLor xml_change (xml2equational_proof xml2name) (xml_return \<circ> Equational_Proof)
        XMLor xml_change (xml2equational_disproof xml2name) (xml_return \<circ> Equational_Disproof)
        XMLor xml_change (xml2complexity_proof xml2name) (xml_return \<circ> Complexity_Proof)
        XMLor xml_change (xml2quasi_reductive_proof xml2name) (xml_return \<circ> Quasi_Reductive_Proof)
        XMLor xml_change (xml2conditional_cr_proof xml2name) (xml_return \<circ> Conditional_CR_Proof)
        XMLor xml_change (xml2conditional_ncr_proof xml2name) (xml_return \<circ> Conditional_Non_CR_Proof)
        XMLor xml_singleton ''treeAutomatonClosedProof'' closed_criterion Tree_Automata_Closed_Proof
        XMLor xml_change (xml2ac_termination_proof xml2name) (xml_return \<circ> AC_Termination_Proof)
        XMLor xml_singleton ''ltsTerminationProof'' IA.lts_termination_proof_parser LTS_Termination_Proof
        XMLor xml_singleton ''ltsSafetyProof'' IA.lts_safety_proof_parser LTS_Safety_Proof
        XMLor xml_change (xml2unknown_proof xml2name) (xml_return \<circ> Unknown_Proof)
        XMLor xml_change (xml2unknown_disproof xml2name) (xml_return \<circ> Unknown_Disproof);
      xml_return p
    };
    orig \<leftarrow>? xml_any;
    xml_return (inp, p)
  }"

function plain_name :: "lsymbol xmlt2"
where "plain_name x = (
    xml_change (xml_text ''name'') (xml_return \<circ> UnLab)
    XMLor xml_singleton ''sharp'' plain_name Sharp
    XMLor xml_pair ''labeledSymbol'' plain_name (
        xml_many ''numberLabel'' (xml_nat ''number'') Inl
        XMLor xml_many ''symbolLabel'' plain_name Inr
      )
      (\<lambda>f l. case l of
        Inl ls  \<Rightarrow> Lab f ls
      | Inr ls \<Rightarrow> FunLab f ls)
  ) x"
by (pat_completeness,auto)

termination by (relation "measure size_xml_state", auto)

definition "parse_cert_problem \<equiv>
  parse_xmlfile (
    debug ''0'' ''parsing xml to certification problem''
    (xml2cert_problem (xml_text ''name'') CPF_Parser.plain_name))"

definition xml_symbols :: "string \<Rightarrow> lsymbol list \<Rightarrow> xml" where
  "xml_symbols tagname Fs \<equiv> XML tagname [] []"

definition xml_equation_literal :: "(_,_) equation_literal \<Rightarrow> xml" where
  "xml_equation_literal eql \<equiv>
    case eql of Equation (s,t) \<Rightarrow> XML ''equation'' [] [xml_term s, xml_term t]
              | Inequality (s,t) \<Rightarrow> XML ''inequality'' [] [xml_term s, xml_term t]"
  
primrec xml_input :: "_ \<Rightarrow> xml" where
  "xml_input (Inn_TRS_input Q R S start) = xml_trs_input (Inl Q) R S"
| "xml_input (FP_TRS_input Q R) = xml_trs_input (Inr Q) R []"
| "xml_input (DP_input m P Q R) =
    XML ''dpInput'' [] (
      xml_rules ''trs'' R #
      xml_rules ''dps'' P #
      xml_strategy (Inl Q) @ [XML ''minimal'' [] [XML_text (if m then ''true'' else ''false'')]])"
| "xml_input (COMP_input E R) =
    XML ''completionInput'' [] [xml_rules ''equations'' E, xml_rules ''trs'' R]"
| "xml_input (OCOMP_input E\<^sub>0 E R ro) =
    XML ''orderedCompletionInput'' [] [xml_rules ''equations'' E\<^sub>0, xml_rules ''trs'' R, xml_rules ''equations'' E, XML ''reductionOrder'' [] [xml_redord ro]]"
| "xml_input (EQ_input E eql) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, xml_equation_literal eql]"
| "xml_input (CPX_input Q R S cm cc) =
    XML ''complexityInput'' [] [xml_trs_input (Inl Q) R S, xml_complexity_measure cm, xml_complexity_class cc]"
| "xml_input (CTRS_input ctrs) = xml_ctrs_input ctrs"
| "xml_input (TA_input ta r) = xml_ta_input ta r"
| "xml_input (AC_input r a c) = xml_ac_input r a c"
| "xml_input (LTS_input r) = xml_lts r"
| "xml_input (LTS_safety_input r e) = xml_lts_safety r e"
| "xml_input (Unknown_input u) = XML ''unknownInput'' [] [XML_text u]"

hide_const(open) rstep estep cstep

end
